import asyncio

from typing_extensions import TYPE_CHECKING, Any, AsyncGenerator, NotRequired, TypedDict, cast

from .reflect import logger

if TYPE_CHECKING:
    import socket

from .._run import set_loop_exc_handler
from ..typ import ExitCode, LogLevel


class LoopExcCtx(TypedDict):
    message: str
    exception: NotRequired[BaseException]
    future: NotRequired[asyncio.Future]
    task: NotRequired[asyncio.Task]
    handle: NotRequired[asyncio.Handle]
    protocol: NotRequired[asyncio.Protocol]
    transport: NotRequired[asyncio.Transport]
    socket: NotRequired["socket.socket"]
    asyncgen: NotRequired[AsyncGenerator]


_STRICT_LOOP_LOG = False


def set_loop_exc_log(strict: bool) -> None:
    global _STRICT_LOOP_LOG
    _STRICT_LOOP_LOG = strict


def _log_loop_exception(loop: asyncio.AbstractEventLoop, context: dict[str, Any]) -> None:
    ctx = cast(LoopExcCtx, context)
    with_loop_ctx = {"loop": loop} | ctx
    exc = ctx.get("exception")
    msg = ctx["message"]

    if exc is not None:
        if (
            isinstance(exc, SystemExit)
            and exc.code is not None
            and int(exc.code) == ExitCode.RESTART.value
        ):
            logger.debug("收到重启信号，即将重启...")

        elif "exception was never retrieved" in msg:
            fut = ctx.get("future")
            task = ctx.get("task")
            if _STRICT_LOOP_LOG:
                try:
                    raise exc
                except BaseException:
                    logger.exception(f"从未捕获的异常的回溯栈：{msg}")
            logger.generic_obj(
                f"发现从未捕获的异常（这不一定是错误）：{msg}",
                {"future": fut, "task": task},
                level=LogLevel.ERROR if _STRICT_LOOP_LOG else LogLevel.DEBUG,
            )

        else:
            try:
                raise exc
            except BaseException:
                logger.exception(f"事件循环中抛出预期外的异常：{msg}")
                logger.generic_obj("相关变量信息：", with_loop_ctx, level=LogLevel.ERROR)

    else:
        logger.error(f"事件循环出现预期外的状况：{msg}")
        logger.generic_obj("相关变量信息：", with_loop_ctx, level=LogLevel.ERROR)


set_loop_exc_handler(_log_loop_exception)


def log_exc(exc: BaseException, msg: str, obj: Any = None) -> None:
    """使用 melobot 内部的异常报告日志器报告异常

    :param exc: 异常对象
    :param msg: 日志消息
    :param obj: 需要打印的相关变量，为空时不打印
    """
    try:
        raise exc
    except BaseException:
        logger.exception(msg)
        if obj is not None:
            logger.generic_obj("相关变量信息：", obj, level=LogLevel.ERROR)



import asyncio
from functools import partial
from multiprocessing import SimpleQueue
from pathlib import Path

from melobot._render import get_rich_repr
from melobot.mp import MP_MODULE_NAME, PBox, Process, ProcessPool, ProcessPoolExecutor
from tests.base import *
from tests.mp.mod import simple_test

EMPTY_ENTRY_PATH = Path(__file__).parent.joinpath("mp", "empty.py").resolve()
MOD_PATH = Path(__file__).parent.joinpath("mp", "mod.py").resolve()
SUBMOD_PATH = Path(__file__).parent.joinpath("mp", "submod.py").resolve()
SUBMOD2_PATH = Path(__file__).parent.joinpath("mp", "submod2.py").resolve()
ARGV = ["123", "456"]
SIGNAL_QUEUE = SimpleQueue()
TEST_S = "/abc/123, '123456' <test>(123)45678900012"


async def test_single_process():
    name = f"{test_single_process.__name__}-p"
    p = Process(MOD_PATH, ARGV, PBox(simple_test, entry=MOD_PATH), name=name)
    p.start()
    p.join()
    assert p.exitcode == 0

    p = Process(
        MOD_PATH,
        ARGV,
        PBox(name="test_all", entry=MOD_PATH),
        name=name,
        args=(ARGV, MP_MODULE_NAME),
    )
    p.start()
    p.join()
    assert p.exitcode == 0

    p = Process(MOD_PATH, ARGV, print, name=name, args=(partial(get_rich_repr, TEST_S),))
    p.start()
    p.join()
    assert p.exitcode == 0

    p = Process(
        MOD_PATH,
        ARGV,
        print,
        name=name,
        args=(PBox(get_rich_repr, module="melobot._render"), TEST_S),
    )
    p.start()
    p.join()
    assert p.exitcode == 0


async def test_process_pool():
    func = PBox(name="sync_get_sum", entry=SUBMOD_PATH)
    with ProcessPool(entry=SUBMOD_PATH, argv=ARGV, processes=2) as pool:
        results = pool.map(func, range(1000, 2000, 200))
        assert results == [500500, 720600, 980700, 1280800, 1620900]

        ares = pool.apply_async(func, (2000,))
        res = ares.get(timeout=10)
        assert res == 2001000
        pool.terminate()


async def test_process_pool_executor():
    func = PBox(name="get_rich_str", module="mp.submod2", entry=SUBMOD2_PATH)
    with ProcessPoolExecutor(entry=EMPTY_ENTRY_PATH, max_workers=2) as pool:
        loop = asyncio.get_running_loop()
        res = await asyncio.gather(
            *tuple(loop.run_in_executor(pool, func, TEST_S) for _ in range(10))
        )
        pool.shutdown(wait=False)
    assert res[0][-1] == TEST_S



from __future__ import annotations

from asyncio import create_task, get_running_loop

from typing_extensions import Callable, Iterable, NoReturn

from ..adapter.base import Event
from ..ctx import BotCtx, EventCompletion, FlowCtx, FlowRecord, FlowRecords
from ..ctx import FlowRecordStage as RecordStage
from ..ctx import FlowStatus, FlowStore
from ..di import DependNotMatched, inject_deps
from ..exceptions import FlowError
from ..log.report import log_exc
from ..typ.base import AsyncCallable, SyncOrAsyncCallable
from ..utils.base import to_async
from ..utils.common import get_obj_name
from .graph import DAGMapping

_FLOW_CTX = FlowCtx()


def node(func: SyncOrAsyncCallable[..., bool | None]) -> FlowNode:
    """处理结点装饰器，将当前异步可调用对象装饰为一个处理结点"""
    return FlowNode(func)


def no_deps_node(func: SyncOrAsyncCallable[..., bool | None]) -> FlowNode:
    """与 :func:`node` 类似，但是不自动为结点标记依赖注入。

    需要后续使用 :func:`.inject_deps` 手动标记依赖注入，
    这适用于某些对处理结点进行再装饰的情况
    """
    return FlowNode(func, no_deps=True)


class FlowNode:
    """处理流结点"""

    def __init__(self, func: SyncOrAsyncCallable[..., bool | None], no_deps: bool = False) -> None:
        self.name = get_obj_name(func, otype="callable")
        self.processor: AsyncCallable[..., bool | None] = (
            to_async(func) if no_deps else inject_deps(func)
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name})"

    async def process(self, flow: Flow, completion: EventCompletion) -> None:
        event = completion.event
        try:
            status = _FLOW_CTX.get()
            records, store = status.records, status.store
        except _FLOW_CTX.lookup_exc_cls:
            records, store = FlowRecords(), FlowStore()

        with _FLOW_CTX.unfold(FlowStatus(flow, self, True, completion, records, store)):
            try:
                records.append(FlowRecord(RecordStage.NODE_START, flow.name, self.name, event))

                try:
                    ret = await self.processor()
                    records.append(FlowRecord(RecordStage.NODE_FINISH, flow.name, self.name, event))
                except DependNotMatched as e:
                    ret = False
                    records.append(
                        FlowRecord(
                            RecordStage.DEPENDS_NOT_MATCH,
                            flow.name,
                            self.name,
                            event,
                            f"Real({e.real_type}) <=> Annotation({e.hint})",
                        )
                    )

                if ret in (None, True) and _FLOW_CTX.get().next_valid:
                    await nextn()

            except FlowContinued:
                await nextn()


class Flow:
    """处理流

    :ivar str name: 处理流的标识
    """

    def __init__(
        self,
        name: str,
        *edge_maps: Iterable[Iterable[FlowNode] | FlowNode],
        priority: int = 0,
        guard: SyncOrAsyncCallable[[Event], bool | None] | None = None,
    ) -> None:
        """初始化处理流

        :param name: 处理流的标识
        :param edge_maps: 边映射，遵循 melobot 的 graph edges 表示方法
        :param priority: 处理流的优先级
        :param guard: 守卫函数。在处理流运行前调用，返回 `True` 不再继续运行处理流。默认不启用
        """
        self.name = name
        self.graph = DAGMapping[FlowNode](name, *edge_maps)
        self.priority = priority

        self._active = True
        self._guard = to_async(guard) if guard is not None else None

    @staticmethod
    def from_graph(
        name: str,
        graph: DAGMapping[FlowNode],
        priority: int = 0,
        guard: SyncOrAsyncCallable[[Event], bool | None] | None = None,
    ) -> Flow:
        f = Flow(name, priority=priority, guard=guard)
        f.graph = graph
        return f

    def __repr__(self) -> str:
        output = (
            f"{self.__class__.__name__}(name={self.name}, active={self._active}"
            f", pri={self.priority}, nums={len(self.graph)}"
        )

        if len(self.graph):
            output += f", starts:{len(self.graph.starts)})"
        else:
            output += ")"
        return output

    def update_priority(self, priority: int) -> None:
        """更新处理流优先级

        :param priority: 新优先级
        """
        BotCtx().get()._dispatcher.update(priority, self)

    def dismiss(self) -> None:
        """停用处理流

        停用后将无法处理任何新事件，也无法再次恢复使用
        """
        self._active = False

    def is_active(self) -> bool:
        """判断处理流是否处于可用状态

        :return: 是否可用
        """
        return self._active

    def set_guard(self, guard: SyncOrAsyncCallable[[Event], bool | None]) -> None:
        """设置或重设守卫函数

        :param guard: 守卫函数
        """
        self._guard = to_async(guard) if guard is not None else None

    def link(self, flow: Flow, priority: int | None = None) -> Flow:
        """连接另一处理流返回新处理流，并设置新优先级

        新处理流守卫函数为空，使用 :meth:`set_guard` 自行添加

        :param flow: 连接的新流
        :param priority: 新优先级，若为空，则使用两者中较小的优先级
        :return: 新的处理流
        """
        name = f"{self.name} ~ {flow.name}"
        new_flow = Flow.from_graph(
            name,
            self.graph.link(flow.graph, name),
            priority=priority if priority else min(self.priority, flow.priority),
        )
        return new_flow

    def start(self, node: FlowNode) -> FlowNode:
        """设置处理流起始结点的装饰器

        :param node: 起始结点
        :return: 起始结点
        """
        self.graph.add(node, None)
        return node

    def after(self, node: FlowNode) -> Callable[[FlowNode], FlowNode]:
        """在处理流某一参照结点后，添加新结点的装饰器函数

        :param node: 参照结点
        :return: 被装饰的结点
        """

        def after_wrapped(next_node: FlowNode) -> FlowNode:
            self.graph.add(node, next_node)
            return next_node

        return after_wrapped

    def before(self, node: FlowNode) -> Callable[[FlowNode], FlowNode]:
        """在处理流某一参照结点前，添加新结点的装饰器函数

        :param node: 参照结点
        :return: 被装饰的结点
        """

        def before_wrapped(pre_node: FlowNode) -> FlowNode:
            self.graph.add(pre_node, node)
            return pre_node

        return before_wrapped

    def merge(self, *nodes: FlowNode) -> Callable[[FlowNode], FlowNode]:
        """将处理流某几结点的控制流，在当前结点合并的装饰器函数

        :return: 被装饰的结点
        """

        def merge_wrapped(next_node: FlowNode) -> FlowNode:
            for node in nodes:
                self.graph.add(node, next_node)
            return next_node

        return merge_wrapped

    def fork(self, *nodes: FlowNode) -> Callable[[FlowNode], FlowNode]:
        """将处理流当前结点的控制流，分流到某几个结点的装饰器函数

        :return: 被装饰的结点
        """

        def fork_wrapped(pre_node: FlowNode) -> FlowNode:
            for node in nodes:
                self.graph.add(pre_node, node)
            return pre_node

        return fork_wrapped

    async def _handle(self, event: Event) -> None:
        fut = get_running_loop().create_future()
        create_task(self._run(EventCompletion(event, fut, self)))
        await fut

    async def _run(self, completion: EventCompletion) -> None:
        if self._guard is not None:
            try:
                if not await self._guard(completion.event):
                    return self._try_set_completed(completion)
            except Exception as e:
                log_exc(
                    e,
                    msg=f"事件处理流 {self.name} 守卫函数发生异常",
                    obj={
                        "event_id": completion.event.id,
                        "event": completion.event,
                        "guard": self._guard,
                    },
                )
                return self._try_set_completed(completion)

        starts = self.graph.starts
        if not len(starts):
            return self._try_set_completed(completion)

        event = completion.event
        try:
            status = _FLOW_CTX.get()
            records, store = status.records, status.store
        except _FLOW_CTX.lookup_exc_cls:
            records, store = FlowRecords(), FlowStore()

        with _FLOW_CTX.unfold(FlowStatus(self, starts[0], True, completion, records, store)):
            try:
                if not self.graph._verified:
                    self.graph.verify()

                records.append(FlowRecord(RecordStage.FLOW_START, self.name, starts[0].name, event))

                idx = 0
                while idx < len(starts):
                    try:
                        await starts[idx].process(self, completion)
                        idx += 1
                    except FlowRewound:
                        pass

                records.append(
                    FlowRecord(RecordStage.FLOW_FINISH, self.name, starts[0].name, event)
                )

            except FlowBroke:
                pass

            except Exception as e:
                log_exc(
                    e,
                    msg=f"事件处理流 {self.name} 发生异常",
                    obj={
                        "event_id": event.id,
                        "event": event,
                        "completion": completion.__dict__,
                        "cur_flow": self,
                    },
                )

            finally:
                self._try_set_completed(completion)

    def _try_set_completed(self, completion: EventCompletion) -> None:
        if (
            completion.owner_flow is self
            and not completion.under_session
            and not completion.completed.done()
        ):
            completion.completed.set_result(None)


class _FlowSignal(BaseException): ...


class FlowBroke(_FlowSignal): ...


class FlowContinued(_FlowSignal): ...


class FlowRewound(_FlowSignal): ...


async def nextn() -> None:
    """运行下一处理结点（在处理流中使用）"""
    try:
        status = _FLOW_CTX.get()
        nexts = status.flow.graph[status.node].nexts
        if not status.next_valid:
            return

        idx = 0
        while idx < len(nexts):
            try:
                await nexts[idx].process(status.flow, status.completion)
                idx += 1
            except FlowRewound:
                pass

    except _FLOW_CTX.lookup_exc_cls:
        raise FlowError("此时不在活动的事件处理流中，无法调用下一处理结点") from None
    finally:
        status.next_valid = False


async def block() -> None:
    """阻止当前事件向更低优先级的处理流传播（在处理流中使用）"""
    status = _FLOW_CTX.get()
    status.records.append(
        FlowRecord(RecordStage.BLOCK, status.flow.name, status.node.name, status.completion.event)
    )
    status.completion.event.spread = False


async def stop() -> NoReturn:
    """立即停止当前处理流（在处理流中使用）"""
    status = _FLOW_CTX.get()
    status.records.append(
        FlowRecord(RecordStage.STOP, status.flow.name, status.node.name, status.completion.event)
    )
    status.records.append(
        FlowRecord(
            RecordStage.NODE_EARLY_FINISH,
            status.flow.name,
            status.node.name,
            status.completion.event,
        )
    )
    status.records.append(
        FlowRecord(
            RecordStage.FLOW_EARLY_FINISH,
            status.flow.name,
            status.node.name,
            status.completion.event,
        )
    )
    raise FlowBroke("事件处理流被安全地提早结束，请无视这个内部工作信号")


async def bypass() -> NoReturn:
    """立即跳过当前处理结点剩下的步骤，运行下一处理结点（在处理流中使用）"""
    status = _FLOW_CTX.get()
    status.records.append(
        FlowRecord(
            RecordStage.BYPASS,
            status.flow.name,
            status.node.name,
            status.completion.event,
        )
    )
    status.records.append(
        FlowRecord(
            RecordStage.NODE_EARLY_FINISH,
            status.flow.name,
            status.node.name,
            status.completion.event,
        )
    )
    raise FlowContinued("事件处理流安全地跳过结点执行，请无视这个内部工作信号")


async def rewind() -> NoReturn:
    """立即重新运行当前处理结点（在处理流中使用）"""
    status = _FLOW_CTX.get()
    status.records.append(
        FlowRecord(
            RecordStage.REWIND,
            status.flow.name,
            status.node.name,
            status.completion.event,
        )
    )
    status.records.append(
        FlowRecord(
            RecordStage.NODE_EARLY_FINISH,
            status.flow.name,
            status.node.name,
            status.completion.event,
        )
    )
    raise FlowRewound("事件处理流安全地重复执行处理结点，请无视这个内部工作信号")


async def flow_to(flow: Flow) -> None:
    """立即进入一个其他处理流（在处理流中使用）"""
    status = _FLOW_CTX.get()
    await flow._run(status.completion)



import importlib.resources
import re
import sys
from contextlib import ExitStack
from inspect import getabsfile
from os import PathLike, listdir, remove
from pathlib import Path
from time import time
from types import ModuleType

from typing_extensions import Any, Callable, Iterable, cast

from .._imp import ALL_EXTS, Importer
from .._lazy import singleton
from ..ctx import BotCtx
from ..exceptions import DynamicImpSpecEmpty, PluginAutoGenError, PluginLoadError
from ..log.reflect import logger
from .base import Plugin, PluginPlanner
from .ipc import AsyncShare, SyncShare


def plugin_get_attr(p_name: str, name: str, excepts: Iterable[str]) -> Any:
    if name in excepts or name.startswith("_"):
        raise AttributeError
    obj = BotCtx().get().ipc_manager.get(p_name, name)
    if obj.static:
        return obj.get()
    return obj


_AUTOGEN_COMMENT = r"""# This file is @generated by melobot cli.
# It is not intended for manual editing.
"""


class PluginInitHelper:
    _BASE_INIT_PY_STR = (
        importlib.resources.files(__package__)
        .joinpath("__init__.template.py")
        .read_text(encoding="utf-8")
    )

    @staticmethod
    def _get_init_py_str(p_planner: PluginPlanner) -> str:
        output = re.sub(
            r"_VAR(\d+)",
            lambda matched: f"_{int(time()):#x}{matched.group(1)}",
            PluginInitHelper._BASE_INIT_PY_STR,
        )

        output = f'{output}__version__ = "{p_planner.version}"\n'
        if p_planner.info.author != "":
            output = f'{output}__author__ = "{p_planner.info.author}"\n'
        if p_planner.info.desc != "":
            output = f'"""\n{p_planner.info.desc}\n"""\n{output}'
            output = f'{output}__doc__ = "{p_planner.info.desc}"\n'
        return output

    @staticmethod
    def _get_init_pyi_str(
        plugin_dir: Path,
        shares: Iterable[AsyncShare | SyncShare],
        funcs: Iterable[Callable],
    ) -> str:
        refs: dict[str, list[str]] = {}
        varnames: list[str] = []
        autoget_varnames: list[str] = []

        for share in shares:
            if share.static:
                autoget_varnames.append(share.name)

            share_located = Path(share.__obj_file__)
            try:
                share_abs_located = share_located.resolve(strict=True)
            except Exception as e:
                raise PluginAutoGenError(
                    f"无法定位共享对象所在的模块位置，对应插件：{plugin_dir}"
                ) from e
            else:
                if plugin_dir.resolve(strict=True) not in share_abs_located.parents:
                    raise PluginAutoGenError(
                        f"共享对象 {share} 不在插件目录下定义，对应插件：{plugin_dir}"
                    )
                parts1 = share_abs_located.parts

            parts2 = plugin_dir.parts
            imp_parts = list(parts1[len(parts2) :])
            imp_parts[-1] = imp_parts[-1].rstrip(".py")

            mod = Importer.import_mod(share.__obj_module__, share_located.parent)
            for k in dir(mod):
                v = getattr(mod, k)
                if v is share:
                    share_varname = k
                    break
            else:
                raise PluginAutoGenError(
                    f"共享对象 {share} 在所属模块的全局作用域不存在引用，生成 .pyi 失败，"
                    f"对应插件：{plugin_dir}"
                )

            li = refs.setdefault(f".{'.'.join(imp_parts)}", [])
            if share.static:
                li.append(f"{share_varname} as _{share.name}")
            elif share_varname != share.name:
                li.append(f"{share_varname} as {share.name}")
            else:
                li.append(share_varname)
            varnames.append(share.name)

        for func in funcs:
            func_located = Path(getabsfile(func))
            try:
                func_abs_located = func_located.resolve(strict=True)
            except Exception as e:
                raise PluginAutoGenError(
                    f"无法定位导出函数所在模块的位置，对应插件：{plugin_dir}"
                ) from e
            else:
                if plugin_dir.resolve(strict=True) not in func_abs_located.parents:
                    raise PluginAutoGenError(
                        f"导出函数 {func} 不在插件目录下定义，对应插件：{plugin_dir}"
                    )
                parts1 = func_abs_located.parts

            parts2 = plugin_dir.parts
            imp_parts = list(parts1[len(parts2) :])
            imp_parts[-1] = imp_parts[-1].rstrip(".py")
            if func.__name__.startswith("_"):
                raise PluginAutoGenError(
                    f"导出函数 {func} 的名称不能以 _ 开头，对应插件：{plugin_dir}"
                )

            mod = Importer.import_mod(func.__module__, func_located.parent)
            for k in dir(mod):
                v = getattr(mod, k)
                if v is func:
                    break
            else:
                raise PluginAutoGenError(
                    f"导出函数 {func} 在所属模块的全局作用域不存在引用，生成 .pyi 失败，"
                    f"对应插件：{plugin_dir}"
                )
            refs.setdefault(f".{'.'.join(imp_parts)}", []).append(func.__name__)
            varnames.append(func.__name__)

        imp_lines = "\n".join(
            f"from {namespace} import {name}"
            for namespace, obj_names in refs.items()
            for name in obj_names
        )
        gets_str = "\n".join(f"{varname} = _{varname}.get()" for varname in autoget_varnames)
        vars_str = f"__all__ = {repr(tuple(varnames))}"

        if not len(varnames):
            return ""

        return "\n\n".join((imp_lines, gets_str, vars_str)) + "\n"

    @staticmethod
    def run_init(*plugin_dirs: str | PathLike[str], load_depth: int = 1) -> None:
        from ..bot.base import Bot

        tmp_bot = Bot()
        for p_dir in plugin_dirs:
            try:
                p_dir = Path(p_dir).resolve(strict=True)
            except FileNotFoundError:
                raise PluginAutoGenError(f"插件目录不存在。对应插件：{p_dir}") from None

            p_name = p_dir.parts[-1]
            p_conflicts = set(fname.split(".")[0] for fname in listdir(p_dir))
            pinit_path = p_dir.joinpath("__init__.py")
            pinit_typ_path = p_dir.joinpath("__init__.pyi")

            if p_name in sys.modules or p_name in sys.stdlib_module_names:
                raise PluginAutoGenError(
                    f"插件 {p_name} 与 python 内置模块或已加载模块重名，"
                    f"请修改名称（修改插件目录名）。对应插件：{p_dir}"
                )
            if not p_dir.joinpath("__plugin__.py").exists():
                raise PluginAutoGenError(
                    f"插件目录下不存在 __plugin__.py，无法解析。对应插件：{p_dir}"
                )
            if pinit_path.exists():
                remove(pinit_path)
            if pinit_typ_path.exists():
                remove(pinit_typ_path)

            with ExitStack() as ctx_stack:
                ctx_stack.enter_context(BotCtx().unfold(tmp_bot))

                prefix = ".".join(p_dir.parts[-load_depth:])
                p_entry_mod_name = f"{prefix}.__plugin__"
                p_entry_mod = Importer.import_mod(p_entry_mod_name, p_dir)
                for k in dir(p_entry_mod):
                    val = getattr(p_entry_mod, k)
                    if isinstance(val, PluginPlanner):
                        p_planner = val
                        p_shares = p_planner.shares
                        p_funcs = p_planner.funcs
                        break
                else:
                    raise PluginAutoGenError(
                        f"插件的 __plugin__.py 未实例化 {PluginPlanner.__name__} 类，"
                        f"无法解析。对应插件：{p_dir}"
                    )

                for share in p_shares:
                    if share.name in p_conflicts:
                        raise PluginAutoGenError(
                            f"插件的共享对象名 {share.name} 与插件根目录下的文件/目录名重复，"
                            "这将导致导入混淆，请修改共享对象名。"
                            f"对应插件：{p_dir}"
                        )
                for func in p_funcs:
                    if func.__name__ in p_conflicts:
                        raise PluginAutoGenError(
                            f"插件的导出函数名 {func.__name__} 与插件根目录下的文件/目录名重复，"
                            "这将导致导入混淆，请修改导出函数名"
                            f"对应插件：{p_dir}"
                        )

            try:
                pyi_content = PluginInitHelper._get_init_pyi_str(p_dir, p_shares, p_funcs)
                if pyi_content != "":
                    with open(pinit_typ_path, "w", encoding="utf-8") as fp:
                        fp.write(_AUTOGEN_COMMENT + pyi_content)
                with open(pinit_path, "w", encoding="utf-8") as fp:
                    fp.write(_AUTOGEN_COMMENT + PluginInitHelper._get_init_py_str(p_planner))

            except BaseException:
                if pinit_path.exists():
                    remove(pinit_path)
                if pinit_typ_path.exists():
                    remove(pinit_typ_path)
                raise

            Importer.clear_cache()


P_PLANNER_ATTR = "__plugin_planner__"
P_INFO_ATTR = "__plugin_info__"
_MODULE_EXTS = ALL_EXTS


@singleton
class PluginLoader:
    def __init__(self) -> None:
        self._dir_caches: dict[str, Path] = {}

    def load(
        self, plugin: ModuleType | str | PathLike[str] | PluginPlanner, load_depth: int
    ) -> tuple[Plugin, bool]:
        if isinstance(plugin, PluginPlanner):
            p_name = f"_DynamicPlugin_0x{id(plugin):0x}"
            return self._build_dynamic(p_name, plugin)

        if (
            isinstance(plugin, str)
            and not plugin.endswith(_MODULE_EXTS)
            and not Path(plugin).exists()
        ):
            try:
                tmp_mod = Importer.import_mod(plugin)
                plugin = tmp_mod
            except DynamicImpSpecEmpty:
                pass

        if isinstance(plugin, ModuleType):
            p_dir = self._get_mod_plugin_dir(plugin)
        else:
            p_dir = Path(plugin)
        try:
            p_dir.resolve(strict=True)
        except FileNotFoundError:
            raise PluginLoadError(f"插件目录不存在，无法加载。对应插件：{p_dir}") from None

        p_name = p_dir.parts[-1]
        logger.debug(f"尝试加载来自 {p_dir!r} 的插件：{p_name}")
        return self._build_from_dir(p_name, p_dir, load_depth)

    def _build_dynamic(self, p_name: str, planner: PluginPlanner) -> tuple[Plugin, bool]:
        return self._create_plugin(p_name, planner, None)

    def _build_from_dir(self, p_name: str, p_dir: Path, load_depth: int) -> tuple[Plugin, bool]:
        if p_name in sys.stdlib_module_names:
            raise PluginLoadError(
                f"尝试加载的插件 {p_name} 与 Python 内置模块重名，"
                f"请修改名称（修改插件目录名）。对应插件：{p_dir}"
            )
        if not p_dir.joinpath("__plugin__.py").exists():
            raise PluginLoadError(f"插件目录下不存在 __plugin__.py，无法加载。对应插件：{p_dir}")
        if p_name in self._dir_caches and p_dir.resolve() != self._dir_caches[p_name]:
            raise PluginLoadError(
                f"试图加载一个与已加载插件同名的插件，名称：{p_name}，尝试加载的插件：{p_dir}"
            )

        p_mod = Importer.get_cache(p_dir)
        if p_mod is not None:
            p_entry_mod_name = f"{p_mod.__name__}.__plugin__"
        else:
            prefix = ".".join(p_dir.parts[-load_depth:])
            p_entry_mod_name = f"{prefix}.__plugin__"
        entry = Importer.import_mod(p_entry_mod_name, p_dir)

        if not hasattr(entry, P_PLANNER_ATTR):
            for k in dir(entry):
                val = getattr(entry, k)
                if isinstance(val, PluginPlanner):
                    setattr(entry, P_PLANNER_ATTR, val)
                    break
            else:
                raise PluginLoadError(
                    f"__plugin__.py 未实例化 {PluginPlanner.__name__}，无法加载。"
                    f"对应插件：{p_dir}"
                )

        planner = cast(PluginPlanner, getattr(entry, P_PLANNER_ATTR))
        p, is_repeat = self._create_plugin(p_name, planner, entry)
        if not is_repeat:
            self._dir_caches[p_name] = p_dir.resolve()
            setattr(entry, P_INFO_ATTR, planner.info)
        return p, is_repeat

    def _create_plugin(
        self, p_name: str, planner: PluginPlanner, entry: ModuleType | None
    ) -> tuple[Plugin, bool]:
        if planner._built:
            p = planner._plugin
            logger.debug(f"插件 {p.name} 已加载，重复加载将被跳过")
            return p, True

        if entry is not None:
            p_dir = self._get_mod_plugin_dir(entry).resolve()
            self._auto_import(p_name, entry.__name__, p_dir, planner.auto_import)

        planner._pname = p_name
        planner._hook_bus.set_tag(p_name)
        p = planner._plugin = Plugin(planner)
        planner._built = True
        return p, False

    def _auto_import(
        self, p_name: str, p_entry_mod_name: str, p_dir: Path, paths: Iterable[str] | bool
    ) -> None:
        if paths is False:
            return
        if paths is True:
            paths = map(str, p_dir.glob("**/*.py"))

        p_mod_name = p_entry_mod_name.rsplit(".", maxsplit=1)[0]
        for path_str in paths:
            ext = next((ext for ext in _MODULE_EXTS if path_str.endswith(ext)), None)
            if ext is None:
                raise PluginLoadError(
                    f"插件 {p_name} 的自动导入列表中，{path_str!r} 不是可加载的模块"
                )

            path = Path(path_str)
            if not path.is_absolute():
                path = p_dir.joinpath(path)
            try:
                path = path.resolve().relative_to(p_dir)
            except ValueError:
                raise PluginLoadError(
                    f"插件 {p_name} 自动导入时，以下路径无法计算相对导入关系：{path_str!r}"
                )

            if path.name == f"__plugin__{ext}":
                continue
            elif path.name == f"__init__{ext}":
                path = path.parent
                parts = path.parts
            else:
                parts = (*path.parts[:-1], path.parts[-1].removesuffix(ext))

            mod_name = f"{p_mod_name}.{'.'.join(parts)}"
            # 恢复绝对路径，准备加载
            path = p_dir.joinpath(path).resolve()
            Importer.import_mod(mod_name, path.parent.as_posix())

    def _get_mod_plugin_dir(self, mod: ModuleType) -> Path:
        if mod.__file__ is None:
            p_dir = Path(mod.__path__[0])
        else:
            p_dir = Path(mod.__file__).parent
        return p_dir



from __future__ import annotations

import collections
import io
import os
import sys
from functools import lru_cache
from pathlib import Path
from types import FrameType, TracebackType

from typing_extensions import (
    TYPE_CHECKING,
    Any,
    Generator,
    Iterable,
    Iterator,
    Literal,
    Mapping,
    TypeAlias,
    cast,
    overload,
)

# 使用绝对导入保证子进程导入正确
from melobot._lazy import singleton

if TYPE_CHECKING:
    import rich.console
    import rich.highlighter
    from better_exceptions import ExceptionFormatter
    from rich.style import Style

_ORIGINAL_EXC_HOOK = sys.excepthook

# 使用 singleton 做 lazyload 优化


@singleton
def _get_tmp_console() -> "rich.console.Console":
    import rich.console

    return rich.console.Console(file=_TMP_CONSOLE_IO, record=True, color_system="256")


@singleton
def _get_exc_console() -> "rich.console.Console":
    import rich.console

    return rich.console.Console(file=sys.stderr, color_system="256")


@singleton
def _get_repr_highlighter() -> "rich.highlighter.Highlighter":
    import rich.highlighter

    return rich.highlighter.ReprHighlighter()


@singleton
def _get_style_highlighter_type() -> Any:
    from rich.highlighter import Highlighter

    if TYPE_CHECKING:
        from rich.text import Text

    class _StyleHighlighter(Highlighter):
        def __init__(self, style: "Style" | None) -> None:
            super().__init__()
            self.style = style

        def highlight(self, text: "Text") -> None:
            if self.style:
                text.stylize(self.style)

    return _StyleHighlighter


@singleton
def _get_exc_fmtter() -> "ExceptionFormatter":
    import better_exceptions
    from better_exceptions.formatter import ExceptionFormatter

    better_exceptions.SUPPORTS_COLOR = True
    better_exceptions.color.SUPPORTS_COLOR = True
    better_exceptions.formatter.SUPPORTS_COLOR = True
    # 修复在 windows powershell 显示错误的 bug
    better_exceptions.encoding.ENCODING = sys.stdout.encoding
    better_exceptions.formatter.ENCODING = sys.stdout.encoding

    class ExcFmtter(ExceptionFormatter):
        # 以下代码，由 better-exceptions 模块源代码修改而来
        # 原始版权 © 2016 Josh Junon
        # 原始许可：https://github.com/Qix-/better-exceptions/blob/master/LICENSE.txt
        def __init__(self) -> None:
            super().__init__()
            self._pipe_char = "\x00bold bright_black\x01│\x00/\x01"
            self._cap_char = "\x00bold bright_black\x01└\x00/\x01"
            self._hide_internal = False if EXC_SHOW_INTERNAL in os.environ else True
            self._flip = True if EXC_FLIP in os.environ else False
            self._colored = False

        def set_style(self, hide_internal: bool = True, flip: bool = False) -> None:
            self._hide_internal = hide_internal
            self._flip = flip

        def to_unicode(self, val: bytes | str) -> str:
            if isinstance(val, bytes):
                try:
                    return val.decode()
                except UnicodeDecodeError:
                    return val.decode("unicode-escape")
            return val

        def format_traceback_frame(
            self, tb: TracebackType
        ) -> tuple[tuple[str, int, str, str], str]:
            filename, lineno, function, _, color_source, relevant_values = (
                self.get_traceback_information(tb)
            )

            need_style = False
            if len(color_source.strip()):
                need_style = True
            else:
                color_source = ""
            lines = [color_source]

            for i in reversed(range(len(relevant_values))):
                _, col, val = relevant_values[i]
                pipe_cols = [pcol for _, pcol, _ in relevant_values[:i]]
                line = ""
                index = 0
                for pc in pipe_cols:
                    line += (" " * (pc - index)) + self._pipe_char
                    index = pc + 1

                line += "{}{} {}".format((" " * (col - index)), self._cap_char, val)
                lines.append(self._theme["inspect"](line) if self._colored else line)

            if need_style:
                lines[0] = f"\x00bold white\x01{lines[0]}\x00/\x01"
            formatted = "\n    ".join([self.to_unicode(x) for x in lines])
            return (filename, lineno, function, formatted), color_source

        def format_traceback(self, tb: TracebackType | None = None) -> tuple[str, str]:
            omit_last = False
            if not tb:
                try:
                    raise Exception
                except Exception:
                    omit_last = True
                    _, _, tb = sys.exc_info()
                    if tb is None:
                        raise ValueError("异常的回溯栈信息为空，无法格式化")

            frames = []
            final_source = ""
            while tb:
                if omit_last and not tb.tb_next:
                    break
                formatted, colored = self.format_traceback_frame(tb)
                collectable = True

                if self._flip:
                    formatted = (*formatted[:-1], "")
                    colored = ""

                try:
                    path = Path(formatted[0]).resolve(strict=True)
                    path_str = path.as_posix()
                except Exception:
                    pass
                else:
                    if self._hide_internal and _MAIN_PKG_PATH in path.parents:
                        collectable = False
                    else:
                        formatted = (path_str, *formatted[1:])

                if collectable:
                    final_source = colored
                    frames.append(formatted)
                tb = tb.tb_next

            lines = StackSummary.from_list(frames).format()
            return "".join(lines), final_source

        @lru_cache(maxsize=3)
        def format_exception(self, _: Any, exc: BaseException, tb: TracebackType | None) -> str:
            output = "".join(line for line in self._format_exception(exc, tb)).lstrip("\n").rstrip()
            if self._flip:
                output = output.replace("\n\n", "\n")
            return output

        def _format_exception(
            self, value: BaseException, tb: TracebackType | None, seen: set[int] | None = None
        ) -> Generator[str, None, None]:
            exc_type, exc_value, exc_traceback = type(value), value, tb
            if seen is None:
                seen = set()
            seen.add(id(exc_value))

            if exc_value:
                if exc_value.__cause__ is not None and id(exc_value.__cause__) not in seen:
                    for text in self._format_exception(
                        exc_value.__cause__, exc_value.__cause__.__traceback__, seen=seen
                    ):
                        yield text
                    yield "\nThe above exception was the direct cause of the following exception:\n\n"
                elif (
                    exc_value.__context__ is not None
                    and id(exc_value.__context__) not in seen
                    and not exc_value.__suppress_context__
                ):
                    for text in self._format_exception(
                        exc_value.__context__, exc_value.__context__.__traceback__, seen=seen
                    ):
                        yield text
                    yield "\nDuring handling of the above exception, another exception occurred:\n\n"

            if exc_traceback is not None:
                yield "Traceback (most recent call last):\n\n"

            formatted, colored_source = self.format_traceback(exc_traceback)
            formatted = formatted.replace("[", r"\[").replace("\x00", "[").replace("\x01", "]")
            yield formatted
            if not str(value) and exc_type is AssertionError:
                value.args = (colored_source,)

            te = TracebackException(type(value), value, None, compact=True)
            title_str = "".join(te.format_exception_only()).lstrip("\n").rstrip() + "\n"
            title_str = title_str.replace("[", r"\[").replace("\x00", "[").replace("\x01", "]")
            yield title_str

    return ExcFmtter()


_TMP_CONSOLE_IO = io.StringIO()
_HIGH_LIGHTWORDS = ["GET", "POST", "HEAD", "PUT", "DELETE", "OPTIONS", "TRACE", "PATCH"]
_MAIN_PKG_PATH = Path(cast(str, sys.modules["melobot"].__file__)).parent.resolve()


def get_rich_object(
    obj: object,
    max_len: int | None = 2000,
    style: "Style" | None = None,
    no_color: bool = False,
) -> tuple[str, str]:
    import rich.pretty

    if no_color:
        hl = _get_style_highlighter_type()(style=None)
    elif style:
        hl = _get_style_highlighter_type()(style)
    else:
        hl = None

    _get_tmp_console().print(
        rich.pretty.Pretty(
            obj,
            highlighter=hl,
            indent_guides=True,
            max_string=max_len,
            overflow="ignore",
            expand_all=True,
        ),
        crop=False,
    )
    colored_str = _TMP_CONSOLE_IO.getvalue().rstrip("\n")
    _TMP_CONSOLE_IO.seek(0)
    _TMP_CONSOLE_IO.truncate(0)
    return colored_str, _get_tmp_console().export_text().rstrip("\n")


def get_rich_repr(s: str, style: "Style" | None = None, no_color: bool = False) -> tuple[str, str]:
    from rich.text import Text

    if no_color:
        msg = Text(s)
    elif style:
        msg = Text(s, style=style)
    else:
        msg = _get_repr_highlighter()(Text(s))
        msg.highlight_words(_HIGH_LIGHTWORDS, "logging.keyword")

    _get_tmp_console().print(msg)
    colored_str = _TMP_CONSOLE_IO.getvalue()[:-1]
    _TMP_CONSOLE_IO.seek(0)
    _TMP_CONSOLE_IO.truncate(0)
    return colored_str, _get_tmp_console().export_text()[:-1]


def get_rich_str(s: str, markup: bool = True, emoji: bool = False) -> tuple[str, str]:
    _get_tmp_console().print(s, markup=markup, emoji=emoji, crop=False)
    colored_str = _TMP_CONSOLE_IO.getvalue()[:-1]
    _TMP_CONSOLE_IO.seek(0)
    _TMP_CONSOLE_IO.truncate(0)
    return colored_str, _get_tmp_console().export_text()[:-1]


# 从 3.13 开始，traceback 的格式化发生改变，导致 better-exceptions 无法使用
# 因此以下代码，用于在任意版本中兼容同一套 trackback 格式化
# TODO: 在升级最低版本到 3.11 后，考虑对 ExceptionGroup, add_note 的支持
# TODO: 在升级最低版本到 3.11 后，考虑使用 traceback.StackSummary.format_frame_summary 来过滤回溯栈帧，并更新 FrameInfo 和 Traceback

EXC_SHOW_INTERNAL = "MELOBOT_EXC_SHOW_INTERNAL"
EXC_FLIP = "MELOBOT_EXC_FLIP"


"""
以下代码，由 Python 3.10 的 traceback 模块源代码修改而来
原始版权 © 2001-2024 Python 软件基金会
原始许可：https://github.com/python/cpython/blob/main/LICENSE
"""
_RECURSIVE_CUTOFF = 3

_FrameSummaryTuple: TypeAlias = tuple[str, int, str, str | None]


class FrameSummary:
    __slots__ = ("filename", "lineno", "name", "_line", "locals")

    def __init__(
        self,
        filename: str,
        lineno: int | None,
        name: str,
        *,
        lookup_line: bool = True,
        locals: Mapping[str, str] | None = None,
        line: str | None = None,
    ) -> None:

        self.filename = filename
        self.lineno = lineno
        self.name = name
        self._line = line
        if lookup_line:
            self.line
        self.locals = {k: repr(v) for k, v in locals.items()} if locals else None

    def __eq__(self, other: object | tuple) -> bool:
        if isinstance(other, FrameSummary):
            return (
                self.filename == other.filename
                and self.lineno == other.lineno
                and self.name == other.name
                and self.locals == other.locals
            )
        if isinstance(other, tuple):
            return (self.filename, self.lineno, self.name, self.line) == other
        return NotImplemented

    @overload
    def __getitem__(self, pos: Literal[0]) -> str: ...
    @overload
    def __getitem__(self, pos: Literal[1]) -> int: ...
    @overload
    def __getitem__(self, pos: Literal[2]) -> str: ...
    @overload
    def __getitem__(self, pos: Literal[3]) -> str | None: ...
    @overload
    def __getitem__(self, pos: int) -> Any: ...
    @overload
    def __getitem__(self, pos: slice) -> tuple[Any, ...]: ...

    def __getitem__(self, pos: Any) -> Any:
        return (self.filename, self.lineno, self.name, self.line)[pos]

    def __iter__(self) -> Iterator[Any]:
        return iter([self.filename, self.lineno, self.name, self.line])

    def __repr__(self) -> str:
        return "<FrameSummary file {filename}, line {lineno} in {name}>".format(
            filename=self.filename, lineno=self.lineno, name=self.name
        )

    def __len__(self) -> Literal[4]:
        return 4

    @property
    def line(self) -> str | None:
        import linecache

        if self._line is None:
            if self.lineno is None:
                return None
            self._line = linecache.getline(self.filename, self.lineno)
        return self._line.strip()


class StackSummary(list[FrameSummary]):
    @classmethod
    def extract(
        cls,
        frame_gen: Iterable[tuple[FrameType, int]],
        *,
        limit: int | None = None,
        lookup_lines: bool = True,
        capture_locals: bool = False,
    ) -> StackSummary:
        import itertools
        import linecache

        if limit is None:
            limit = getattr(sys, "tracebacklimit", None)
            if limit is not None and limit < 0:
                limit = 0
        if limit is not None:
            if limit >= 0:
                frame_gen = itertools.islice(frame_gen, limit)
            else:
                frame_gen = collections.deque(frame_gen, maxlen=-limit)

        result = cls()
        fnames = set()
        for f, lineno in frame_gen:
            co = f.f_code
            filename = co.co_filename
            name = co.co_name

            fnames.add(filename)
            linecache.lazycache(filename, f.f_globals)
            if capture_locals:
                f_locals = f.f_locals
            else:
                f_locals = None
            result.append(FrameSummary(filename, lineno, name, lookup_line=False, locals=f_locals))
        for filename in fnames:
            linecache.checkcache(filename)
        if lookup_lines:
            for fs in result:
                fs.line
        return result

    @classmethod
    def from_list(cls, a_list: Iterable[FrameSummary | _FrameSummaryTuple]) -> StackSummary:
        result = StackSummary()
        for frame in a_list:
            if isinstance(frame, FrameSummary):
                result.append(frame)
            else:
                filename, lineno, name, line = frame
                result.append(FrameSummary(filename, lineno, name, line=line))
        return result

    def format(self) -> list[str]:
        result = []
        last_file = None
        last_line = None
        last_name = None
        count = 0
        for frame in self:
            if (
                last_file is None
                or last_file != frame.filename
                or last_line is None
                or last_line != frame.lineno
                or last_name is None
                or last_name != frame.name
            ):
                if count > _RECURSIVE_CUTOFF:
                    count -= _RECURSIVE_CUTOFF
                    result.append(
                        f"  [Previous line repeated {count} more "
                        f'time{"s" if count > 1 else ""}]\n\n'
                    )
                last_file = frame.filename
                last_line = frame.lineno
                last_name = frame.name
                count = 0
            count += 1
            if count > _RECURSIVE_CUTOFF:
                continue
            row = []
            row.append(
                f"  \x00blue\x01{frame.filename}\x00/\x01"
                f":\x00cyan\x01{frame.lineno}\x00/\x01"
                f" -> \x00magenta\x01{frame.name}\x00/\x01\n"
            )
            if frame.line:
                row.append("    {}\n\n".format(frame.line.strip()))
            if frame.locals:
                for name, value in sorted(frame.locals.items()):
                    row.append("    {name} = {value}\n\n".format(name=name, value=value))
            result.append("".join(row))
        if count > _RECURSIVE_CUTOFF:
            count -= _RECURSIVE_CUTOFF
            result.append(
                f"  [Previous line repeated {count} more " f'time{"s" if count > 1 else ""}]\n\n'
            )
        return result


class TracebackException:
    def __init__(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType | None,
        *,
        limit: int | None = None,
        lookup_lines: bool = True,
        capture_locals: bool = False,
        compact: bool = False,
        _seen: set[int] | None = None,
    ) -> None:
        is_recursive_call = _seen is not None
        if _seen is None:
            _seen = set()
        _seen.add(id(exc_value))

        self.stack = StackSummary.extract(
            self.walk_tb(exc_traceback),
            limit=limit,
            lookup_lines=lookup_lines,
            capture_locals=capture_locals,
        )
        self.exc_type = exc_type

        self._str = self._some_str(exc_value)
        if exc_type and issubclass(exc_type, SyntaxError):
            exc_value = cast(SyntaxError, exc_value)
            self.filename = exc_value.filename
            lno = exc_value.lineno
            self.lineno = str(lno) if lno is not None else None
            end_lno = exc_value.end_lineno
            self.end_lineno = str(end_lno) if end_lno is not None else None
            self.text = exc_value.text
            self.offset = exc_value.offset
            self.end_offset = exc_value.end_offset
            self.msg = exc_value.msg
        if lookup_lines:
            self._load_lines()
        self.__suppress_context__ = (
            exc_value.__suppress_context__ if exc_value is not None else False
        )

        self.__cause__: TracebackException | None
        self.__context__: TracebackException | None

        if not is_recursive_call:
            queue: list[tuple[TracebackException | None, BaseException | None]] = [
                (self, exc_value)
            ]
            while queue:
                te, e = queue.pop()
                if e and e.__cause__ is not None and id(e.__cause__) not in _seen:
                    cause = TracebackException(
                        type(e.__cause__),
                        e.__cause__,
                        e.__cause__.__traceback__,
                        limit=limit,
                        lookup_lines=lookup_lines,
                        capture_locals=capture_locals,
                        _seen=_seen,
                    )
                else:
                    cause = None

                if compact:
                    need_context = cause is None and e is not None and not e.__suppress_context__
                else:
                    need_context = True
                if (
                    e
                    and e.__context__ is not None
                    and need_context
                    and id(e.__context__) not in _seen
                ):
                    context = TracebackException(
                        type(e.__context__),
                        e.__context__,
                        e.__context__.__traceback__,
                        limit=limit,
                        lookup_lines=lookup_lines,
                        capture_locals=capture_locals,
                        _seen=_seen,
                    )
                else:
                    context = None
                te = cast(TracebackException, te)
                e = cast(BaseException, e)
                te.__cause__ = cause
                te.__context__ = context
                if cause:
                    queue.append((te.__cause__, e.__cause__))
                if context:
                    queue.append((te.__context__, e.__context__))

    @staticmethod
    def walk_tb(tb: TracebackType | None) -> Generator[tuple[FrameType, int], None, None]:
        while tb is not None:
            yield tb.tb_frame, tb.tb_lineno
            tb = tb.tb_next

    @staticmethod
    def _some_str(value: Any) -> str:
        try:
            return str(value)
        except Exception:
            return "<unprintable %s object>" % type(value).__name__

    def _format_final_exc_line(self, etype: str | None, value: str) -> str:
        valuestr = self._some_str(value)
        if value is None or not valuestr:
            line = "%s\n\n" % etype
        else:
            line = "\x00bold red\x01%s\x00/\x01: %s\n\n" % (etype, valuestr)
        return line

    @classmethod
    def from_exception(
        cls,
        exc: BaseException,
        *,
        limit: int | None = None,
        lookup_lines: bool = True,
        capture_locals: bool = False,
        compact: bool = False,
    ) -> TracebackException:
        return cls(
            type(exc),
            exc,
            exc.__traceback__,
            limit=limit,
            lookup_lines=lookup_lines,
            capture_locals=capture_locals,
            compact=compact,
        )

    def _load_lines(self) -> None:
        for frame in self.stack:
            frame.line

    def __eq__(self, other: object) -> bool:
        if isinstance(other, TracebackException):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __str__(self) -> str:
        return self._str

    def format_exception_only(self) -> Generator[str, None, None]:
        if self.exc_type is None:
            yield self._format_final_exc_line(None, self._str)
            return

        stype = self.exc_type.__qualname__
        smod = self.exc_type.__module__
        if smod not in ("__main__", "builtins"):
            if not isinstance(smod, str):
                smod = "<unknown>"
            stype = smod + "." + stype

        if not issubclass(self.exc_type, SyntaxError):
            yield self._format_final_exc_line(stype, self._str)
        else:
            yield from self._format_syntax_error(stype)

    def _format_syntax_error(self, stype: str) -> Generator[str, None, None]:
        filename_suffix = ""
        need_newline = False
        if self.lineno is not None:
            yield f"  \x00blue\x01{self.filename or 'string'}\x00/\x01:\x00cyan\x01{self.lineno}\x00/\x01\n"
            need_newline = True
        elif self.filename is not None:
            filename_suffix = " (\x00blue\x01{}\x00/\x01)".format(self.filename)

        text = self.text
        if text is not None:
            rtext = text.rstrip("\n")
            ltext = rtext.lstrip(" \n\f")
            spaces = len(rtext) - len(ltext)
            yield "    \x00bold white\x01{}\x00/\x01\n".format(ltext)

            if self.offset is not None:
                offset = self.offset
                end_offset = self.end_offset if self.end_offset not in {None, 0} else offset
                end_offset = cast(int, end_offset)
                if offset == end_offset or end_offset == -1:
                    end_offset = offset + 1

                colno = offset - 1 - spaces
                end_colno = end_offset - 1 - spaces
                if colno >= 0:
                    caretspace = ((c if c.isspace() else " ") for c in ltext[:colno])
                    yield "    {}{}".format("".join(caretspace), ("^" * (end_colno - colno)))
        msg = self.msg or "<no detail available>"
        prefix = "\n\n" if need_newline else ""
        yield "{}\x00bold red\x01{}\x00/\x01: {}{}\n".format(prefix, stype, msg, filename_suffix)


# -----------------------------------------------------------------------------


@lru_cache(maxsize=3)
def get_rich_exception(
    exc_type: type[BaseException], exc: BaseException, tb: TracebackType | None
) -> tuple[str, str]:
    lines = _get_exc_fmtter().format_exception(exc_type, exc, tb)  # type: ignore[arg-type]
    colored_str, plain_str = get_rich_str(lines)
    return colored_str, plain_str


def _excepthook(
    exc_type: type[BaseException], exc: BaseException, tb: TracebackType | None
) -> None:
    lines = _get_exc_fmtter().format_exception(exc_type, exc, tb)  # type: ignore[arg-type]
    _get_exc_console().print(lines, markup=True, emoji=False, crop=False)


def install_exc_hook() -> None:
    """安装 melobot 默认异常格式化的钩子

    在 Jupyter 或 IPython 中使用时，需要手动调用该方法安装

    但在普通的脚本环境中，导入时已自动安装
    """

    def ipy_excepthook_closure(ip: Any) -> None:
        # 以下代码，由 rich 模块源代码修改而来
        # 原始版权 © 2020 Will McGugan
        # 原始许可：https://github.com/Textualize/rich/blob/master/LICENSE
        tb_data = {}
        default_showtraceback = ip.showtraceback

        def ipy_show_traceback(*args: Any, **kwargs: Any) -> None:
            nonlocal tb_data
            tb_data = kwargs
            default_showtraceback(*args, **kwargs)

        def ipy_display_traceback(*args: Any, is_syntax: bool = False, **kwargs: Any) -> None:
            nonlocal tb_data
            exc_tuple = ip._get_exc_info()
            tb: TracebackType | None = None if is_syntax else exc_tuple[2]

            compiled = tb_data.get("running_compiled_code", False)
            tb_offset = tb_data.get("tb_offset", 1 if compiled else 0)
            for _ in range(tb_offset):
                if tb is None:
                    break
                tb = tb.tb_next

            _excepthook(exc_tuple[0], exc_tuple[1], tb)
            tb_data = {}

        ip._showtraceback = ipy_display_traceback
        ip.showtraceback = ipy_show_traceback
        ip.showsyntaxerror = lambda *args, **kwargs: ipy_display_traceback(
            *args, is_syntax=True, **kwargs
        )

    try:
        ip = get_ipython()  # type: ignore[name-defined] # noqa: F821
        ipy_excepthook_closure(ip)
    except Exception:
        sys.excepthook = _excepthook


EXC_FMT_FALLBACK = "MELOBOT_EXC_FMT_FALLBACK"
if EXC_FMT_FALLBACK not in os.environ:
    sys.excepthook = _excepthook


def uninstall_exc_hook() -> None:
    """卸载 melobot 默认异常格式化的钩子

    在任意环境中皆可使用
    """
    sys.excepthook = _ORIGINAL_EXC_HOOK


def set_traceback_style(hide_internal: bool = True, flip: bool = False) -> None:
    """melobot 默认异常格式化启用时，设置异常回溯栈的格式化风格

    :param hide_internal: 是否隐藏 melobot 内部的栈帧
    :param flip: 是否折叠异常栈帧，只显示文件和行号，而不是具体行的代码
    """
    _get_exc_fmtter().set_style(hide_internal, flip)



from typing_extensions import Any

# 兼容老版本的导入
from ..typ._enum import LogLevel
from .base import GenericLogger, Logger, NullLogger
from .patch import LazyLogMethod, LoguruPatch, StandardPatch, StructlogPatch, logger_patch
from .reflect import reflect_logger as _reflect_logger
from .reflect import set_global_logger, set_module_logger
from .report import log_exc


def get_logger() -> GenericLogger:
    """获取当前域下可用的日志器

    :return: 日志器
    """
    return _reflect_logger()


def __getattr__(name: str) -> Any:
    if name == "logger":
        return _reflect_logger()
    else:
        raise AttributeError


logger: GenericLogger
"""当前上下文中的日志器"""



from typing_extensions import Any

from ..ctx import BotCtx as _BotCtx
from .base import Bot, BotLifeSpan


def get_bot() -> Bot:
    """获得当前上下文中的 bot 对象

    :return: bot 对象
    """
    return _BotCtx().get()


def __getattr__(name: str) -> Any:
    if name == "bot":
        return get_bot()
    else:
        raise AttributeError


bot: Bot
"""当前上下文中的 bot 对象"""



import sys
import zipimport

if sys.version_info < (3, 11):
    from importlib._bootstrap_external import _NamespaceLoader
else:
    from importlib.machinery import NamespaceLoader as _NamespaceLoader

from importlib.abc import FileLoader, Loader, MetaPathFinder
from importlib.machinery import (
    BYTECODE_SUFFIXES,
    EXTENSION_SUFFIXES,
    SOURCE_SUFFIXES,
    ExtensionFileLoader,
    ModuleSpec,
)
from importlib.machinery import PathFinder as _PathFinder
from importlib.machinery import (
    SourceFileLoader,
    SourcelessFileLoader,
)
from importlib.util import module_from_spec, spec_from_file_location
from os import PathLike
from pathlib import Path
from types import ModuleType

from typing_extensions import Any, Sequence, cast

from ._lazy import singleton
from .exceptions import DynamicImpError, DynamicImpSpecEmpty

# 扩展名的优先级顺序非常重要
ALL_EXTS = tuple(EXTENSION_SUFFIXES + SOURCE_SUFFIXES + BYTECODE_SUFFIXES)
PKG_INIT_FILENAMES = tuple(f"__init__{ext}" for ext in ALL_EXTS)
EMPTY_PKG_TAG = "__melobot_namespace_pkg__"
ZIP_MODULE_TAG = "__melobot_zip_module__"


def _get_file_loaders() -> list[tuple[type[Loader], list[str]]]:
    extensions = (ExtensionFileLoader, EXTENSION_SUFFIXES)
    source = (SourceFileLoader, SOURCE_SUFFIXES)
    bytecode = (SourcelessFileLoader, BYTECODE_SUFFIXES)
    return [extensions, source, bytecode]


class _NestedQuickExit(BaseException): ...


@singleton
class SpecFinder(MetaPathFinder):
    def find_spec(
        self,
        fullname: str,
        paths: Sequence[str] | None,
        target: ModuleType | None = None,
        cache: bool = True,
    ) -> ModuleSpec | None:
        if fullname.startswith(_IMP_FALLBACKS):
            return None

        if paths is None:
            paths = sys.path
        if "." in fullname:
            *_, name = fullname.split(".")
        else:
            name = fullname

        mod_path: Path | None = None
        submod_locs: list[str] | None = None

        # 模块查找的优先级，遵循 PEP420: https://peps.python.org/pep-0420/#specification
        try:
            for entry in paths:
                entry_path = Path(entry)
                dir_path = entry_path.joinpath(name)
                pkg_init_paths = tuple(
                    dir_path.joinpath(filename) for filename in PKG_INIT_FILENAMES
                )

                # 带有 __init__.* 的包优先
                for pkg_init_path in pkg_init_paths:
                    if pkg_init_path.exists():
                        mod_path = pkg_init_path
                        submod_locs = [dir_path.resolve().as_posix()]
                        raise _NestedQuickExit

                # 其次是各种可加载的文件
                for ext in ALL_EXTS:
                    _mod_path = entry_path.joinpath(f"{name}{ext}")
                    if _mod_path.exists():
                        mod_path = _mod_path
                        submod_locs = None
                        raise _NestedQuickExit

                # 再次是 zip 文件导入
                if entry_path.suffix == ".zip" and entry_path.exists():
                    zip_importer = zipimport.zipimporter(entry_path.as_posix())
                    spec = zip_importer.find_spec(fullname, target)

                    if spec is not None:
                        if spec.origin is None or spec.origin == "":
                            raise DynamicImpError(
                                f"zip file from {entry_path}, module named {fullname} from {target}, "
                                "failed to get spec origin",
                                name=fullname,
                                path=entry_path.as_posix(),
                            )
                        if spec.loader is None:
                            raise DynamicImpError(
                                f"zip file from {entry_path}, module named {fullname} from {target}, "
                                "spec has no loader",
                                name=fullname,
                                path=entry_path.as_posix(),
                            )

                        spec.loader = ModuleLoader(
                            fullname, Path(spec.origin).resolve(), cache, spec.loader
                        )
                        setattr(spec, ZIP_MODULE_TAG, True)
                        return spec

                # 没有 __init__.* 的包最后查找，spec 设置为与内置导入兼容的命名空间包格式
                if dir_path.exists() and dir_path.is_dir():
                    dir_path_str = dir_path.resolve().as_posix()
                    loader = cast(
                        Loader,
                        _NamespaceLoader(
                            fullname,
                            [dir_path_str],
                            _PathFinder._get_spec,  # type: ignore[attr-defined]
                        ),
                    )
                    spec = spec_from_file_location(
                        fullname,
                        dir_path_str,
                        loader=ModuleLoader(fullname, dir_path, cache, loader),
                        submodule_search_locations=loader._path,  # type: ignore[attr-defined]
                    )

                    if spec is None:
                        raise DynamicImpSpecEmpty(
                            f"package from {dir_path} without __init__ file create spec failed",
                            name=fullname,
                            path=dir_path.as_posix(),
                        )

                    spec.has_location = False
                    spec.origin = None
                    setattr(spec, EMPTY_PKG_TAG, True)
                    return spec

        except _NestedQuickExit:
            pass

        if mod_path is None and submod_locs is None:
            return None

        mod_path = cast(Path, mod_path).resolve()
        return spec_from_file_location(
            fullname,
            mod_path.as_posix(),
            loader=ModuleLoader(fullname, mod_path, cache),
            submodule_search_locations=submod_locs,
        )


@singleton
class ModuleCacher:
    def __init__(self) -> None:
        self._caches: dict[Path, ModuleType] = {}
        self.sync_all_cache()

    # 此处不过多考虑与 sys.modules 的同步问题,
    # 喜爱 hack sys.modules 以至于不兼容的模块，让用户使用导入回退来处理

    # 严格来说，导入缓存应该考虑线程安全问题。
    # 但无论是内置 import 行为（import 语句，__import__ 等），
    # 还是 mb 的动态导入，只要不在非主线程中运行，实际上都不会存在问题。
    # 只需鼓励用户只进行主线程导入即可，这也是十分合理的编程规范。
    # 极少的多线程导入用例，不值得付出加锁的性能代价

    def has_cache(self, mod: ModuleType) -> bool:
        return mod in self._caches.values()

    def get_cache(self, path: Path) -> ModuleType | None:
        # 对应有 __init__.* 的包模块
        if path.parts[-1] in PKG_INIT_FILENAMES:
            path = path.parent
        return self._caches.get(path)

    def set_cache(self, name: str, mod: ModuleType) -> None:
        if (
            mod in self._caches.values()
            or name in sys.stdlib_module_names
            or not hasattr(mod, "__file__")
        ):
            return

        # __file__ 存在且不为空，可能包或任意可被加载的文件
        if mod.__file__ is not None:
            fp = Path(mod.__file__)
            if fp.parts[-1] in PKG_INIT_FILENAMES:
                self._caches[fp.parent] = mod
            else:
                self._caches[fp] = mod
        # 若 __file__ 为空则有 __path__，对应无 __init__.* 的包
        else:
            self._caches[Path(mod.__path__[0])] = mod

    def rm_cache(self, path: Path) -> None:
        # 对应有 __init__.* 的包模块
        if path.parts[-1] in PKG_INIT_FILENAMES:
            path = path.parent
        self._caches.pop(path)

    def sync_all_cache(self) -> None:
        self._caches.clear()
        for name, mod in sys.modules.items():
            self.set_cache(name, mod)


class ModuleLoader(Loader):
    def __init__(
        self,
        fullname: str,
        fp: Path,
        cache: bool,
        inner_loader: Loader | None = None,
    ) -> None:
        super().__init__()
        # 避免在 self.__getattr__() 反射时重名
        self.mb_cacher = ModuleCacher()
        self.mb_fullname = fullname
        self.mb_fp = fp
        self.mb_cache = cache
        self.mb_inner_loader: Loader | None = inner_loader

        if inner_loader is not None:
            return

        for loader_cls, suffixes in _get_file_loaders():
            if fp.resolve().suffix in suffixes:
                loader_cls = cast(type[FileLoader], loader_cls)
                loader = loader_cls(fullname, fp.as_posix())
                self.mb_inner_loader = loader
                break

    def create_module(self, spec: ModuleSpec) -> ModuleType | None:
        mod = None
        if self.mb_cache:
            mod = self.mb_cacher.get_cache(self.mb_fp)
        if mod is not None and mod.__name__ not in sys.modules:
            # 模块有缓存，但不在 sys.modules 中，说明外部手动移除。
            # 此时要同步删除缓存条目，让行为符合外部预期。
            # 对 sys.modules 的其他修改不需要兼容（极少出现），实在需要就用导入回退
            mod = None
            self.mb_cacher.rm_cache(self.mb_fp)
        if mod is None and self.mb_inner_loader is not None:
            mod = self.mb_inner_loader.create_module(spec)
        return mod

    def exec_module(self, mod: ModuleType) -> None:
        if not self.mb_cacher.has_cache(mod) and self.mb_inner_loader is not None:
            # 遵循先记录原则，防止 exec_module 发起的某些递归导入出现错误
            if self.mb_cache:
                sys.modules[self.mb_fullname] = mod

            try:
                # 设置为与内置导入机制兼容的模式
                if hasattr(mod.__spec__, EMPTY_PKG_TAG):
                    mod.__file__ = None
                self.mb_inner_loader.exec_module(mod)
            except BaseException:
                try:
                    if self.mb_cache:
                        del sys.modules[self.mb_fullname]
                except KeyError:
                    pass
                raise

        # 若 inner_loader 为空，则是纯粹的命名空间包（没有 __init__.* 的包模块）
        # 也就不需要任何实质性的 exec_module 过程
        if self.mb_cache:
            self.mb_cacher.set_cache(self.mb_fullname, mod)

    def __getattr__(self, name: str) -> Any:
        # 直接使用反射，而不是更复杂的继承方案
        # inner_loader 实现了必要的 内省接口、importlib.resources 接口
        if self.mb_inner_loader is not None:
            return getattr(self.mb_inner_loader, name)
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")


_IMP_FALLBACKS: tuple[str, ...] = ()
_IMP_FALLBACK_SET: set[str] = set()


class Importer:
    @staticmethod
    def import_mod(
        name: str, path: str | PathLike[str] | None = None, cache: bool = True
    ) -> ModuleType:
        """动态导入一个模块，非线程安全

        :param name: 模块名
        :param path: 在何处查找模块，为空则按照默认规则查找
        :param cache: 是否在 `sys.modules` 中缓存模块
        :return: 模块
        """
        if name.startswith(_IMP_FALLBACKS):
            raise DynamicImpError(
                f"模块 {name} 被标记为回退默认导入机制，无法使用动态导入",
                name=name,
                path=str(path),
            )

        if path is not None:
            try:
                sep = name.rindex(".")
            except ValueError:
                pass
            else:
                Importer.import_mod(name[:sep], Path(path).parent)

        if cache and name in sys.modules:
            return sys.modules[name]

        spec = SpecFinder().find_spec(name, (str(path),) if path is not None else None, cache=cache)
        if spec is None:
            raise DynamicImpSpecEmpty(
                f"名为 {name} 的模块无法加载，指定的位置：{path}",
                name=name,
                path=str(path),
            )

        mod = module_from_spec(spec)
        if spec.loader is None:
            raise DynamicImpError(
                f"module named {name} and path from {path} has no loader", name=name, path=str(path)
            )
        spec.loader.exec_module(mod)
        return mod

    @staticmethod
    def add_fallback(*names: str) -> None:
        global _IMP_FALLBACKS
        for name in names:
            _IMP_FALLBACK_SET.add(name)
        _IMP_FALLBACKS = tuple(_IMP_FALLBACK_SET)

    @staticmethod
    def clear_cache() -> None:
        ModuleCacher().sync_all_cache()

    @staticmethod
    def get_cache(path: Path) -> ModuleType | None:
        return ModuleCacher().get_cache(path)


sys.meta_path.insert(0, SpecFinder())

# 兼容 pkg_resources 的资源获取操作
# 但此模块于 3.12 删除，因此前向版本不再兼容
if sys.version_info < (3, 12):
    # 部分构建（例如 uv 发行的构建）可能已经缺失 pkg_resources
    try:
        import pkg_resources  # type: ignore[import-untyped]
    except ModuleNotFoundError:
        pass
    else:

        def _union_provider(mod: Any) -> pkg_resources.NullProvider:
            if hasattr(mod.__spec__, ZIP_MODULE_TAG):
                return pkg_resources.ZipProvider(mod)
            return pkg_resources.DefaultProvider(mod)

        pkg_resources.register_loader_type(ModuleLoader, cast(type, _union_provider))
        pkg_resources.register_namespace_handler(SpecFinder, pkg_resources.file_ns_handler)


def add_import_fallback(*names: str) -> None:
    """添加未导入模块或包的导入回退，非线程安全

    绝大多数情况下，melobot 都能处理好导入行为。
    但在极少数情况下，某些包或模块可能需要回退到默认的导入机制

    使用此方法，将模块名以 `names` 起始的模块或包标记为需要回退

    :param names: 需要回退的模块或包名称的起始字符串
    """
    Importer.add_fallback(*names)



from typing_extensions import TYPE_CHECKING

# 让 melobot 启用内置异常格式化器
from . import _render
from ._lazy import lazy_import as _lazy_import
from ._lazy import lazy_load

# TODO: 考虑在最低支持 3.11 后，使用 asyncio.TaskGroup 替代多数任务生成场景

if TYPE_CHECKING:
    from ._meta import MetaInfo, __version__
    from ._render import install_exc_hook, set_traceback_style, uninstall_exc_hook
    from .adapter import Action, Adapter, Echo, Event
    from .adapter.generic import send_image, send_text
    from .bot import Bot, get_bot
    from .ctx import Context
    from .di import Depends
    from .handle import (
        Flow,
        FlowDecorator,
        FlowStore,
        node,
        on_command,
        on_contain_match,
        on_end_match,
        on_event,
        on_full_match,
        on_regex_match,
        on_start_match,
        on_text,
        rewind,
        stop,
    )
    from .log import GenericLogger, Logger, get_logger
    from .plugin import AsyncShare, PluginInfo, PluginLifeSpan, PluginPlanner, SyncShare
    from .session import DefaultRule, Rule, Session, SessionStore, enter_session, suspend
    from .typ._enum import LogicMode, LogLevel
else:
    _lazy_import(
        globals(),
        map={
            "._meta": ("MetaInfo", "__version__"),
            "._render": ("install_exc_hook", "set_traceback_style", "uninstall_exc_hook"),
            ".adapter": ("Action", "Adapter", "Echo", "Event"),
            ".adapter.generic": ("send_image", "send_text"),
            ".bot": ("Bot", "get_bot"),
            ".ctx": ("Context",),
            ".di": ("Depends",),
            ".handle": (
                "Flow",
                "FlowDecorator",
                "FlowStore",
                "node",
                "on_command",
                "on_contain_match",
                "on_end_match",
                "on_event",
                "on_full_match",
                "on_regex_match",
                "on_start_match",
                "on_text",
                "rewind",
                "stop",
            ),
            ".log": ("GenericLogger", "Logger", "get_logger"),
            ".plugin": ("AsyncShare", "PluginInfo", "PluginLifeSpan", "PluginPlanner", "SyncShare"),
            ".session": (
                "DefaultRule",
                "Rule",
                "Session",
                "SessionStore",
                "enter_session",
                "suspend",
            ),
            ".typ._enum": ("LogicMode", "LogLevel"),
        },
        deprecations={},
    )
# 让 melobot 开始接管导入机制
from ._imp import ALL_EXTS as MODULE_EXTS
from ._imp import add_import_fallback



from ._enum import ExitCode, LogicMode, LogLevel, VoidType
from .base import AsyncCallable, P, SyncOrAsyncCallable, T, T_co, U, V, is_subhint, is_type
from .cls import BetterABC, BetterABCMeta, SingletonBetterABCMeta, SingletonMeta, abstractattr



from asyncio import Lock
from functools import partial, wraps

from typing_extensions import Any, Callable, Iterable, Sequence, cast

from ..adapter.model import Event, TextEvent
from ..ctx import FlowCtx, ParseArgsCtx
from ..di import inject_deps
from ..session.base import enter_session
from ..session.option import DefaultRule, Rule
from ..typ._enum import LogicMode
from ..typ.base import AsyncCallable, SyncOrAsyncCallable
from ..utils.check import Checker, checker_join
from ..utils.common import get_obj_name
from ..utils.match import (
    ContainMatcher,
    EndMatcher,
    FullMatcher,
    Matcher,
    RegexMatcher,
    StartMatcher,
)
from ..utils.parse import AbstractParseArgs, CmdArgFormatter, CmdParser, Parser
from .base import Flow, no_deps_node


class FlowDecorator:
    def __init__(
        self,
        checker: Checker | None | SyncOrAsyncCallable[[Event], bool] = None,
        matcher: Matcher | None = None,
        parser: Parser | None = None,
        priority: int = 0,
        block: bool = False,
        temp: bool = False,
        decos: Sequence[Callable[[Callable], Callable]] | None = None,
        rule: Rule[Event] | None = None,
    ) -> None:
        """处理流装饰器

        :param checker: 检查器
        :param matcher: 匹配器（指定匹配器，需要先验证已经是文本事件）
        :param parser: 解析器（指定解析器，需要先验证已经是文本事件）
        :param priority: 优先级
        :param block: 是否阻断向低优先级传播
        :param temp: 是否临时使用（处理一次事件后停用）
        :param decos: 装饰器组
        :param rule: 会话规则
        """
        self.checker: Checker | None
        if callable(checker):
            self.checker = Checker.new(checker)
        else:
            self.checker = checker

        self.matcher = matcher
        self.parser = parser

        self._priority = priority
        self._block = block
        self._decos = decos
        self._rule = rule

        self._temp = temp
        self._invalid = False
        self._lock = Lock()
        self._flow: Flow

    def __call__(self, func: SyncOrAsyncCallable[..., bool | None]) -> Flow:
        func = inject_deps(func)
        if self._decos is not None:
            for deco in reversed(self._decos):
                func = deco(func)

        n = no_deps_node(wraps(func)(partial(self._flow_wrapped, func)))
        n.name = get_obj_name(func, otype="callable")
        self._flow = Flow(
            f"OneBotV11Flow[{n.name}]", (n,), priority=self._priority, guard=self._guard
        )
        return self._flow

    async def _guard(self, event: Event) -> bool:
        if self.checker:
            status = await self.checker.check(event)
            if not status:
                return False

        if self.matcher:
            event = cast(TextEvent, event)
            status = await self.matcher.match(event.text)
            if not status:
                return False

        return True

    async def _flow_wrapped(self, func: AsyncCallable[..., bool | None]) -> bool | None:
        if self._invalid:
            self._flow.dismiss()
            return None

        event = FlowCtx().get().completion.event
        if not self._temp:
            passed, args = await self._parse(event)
            if not passed:
                return None
            return await self._process(func, event, args)

        async with self._lock:
            if self._invalid:
                self._flow.dismiss()
                return None

            passed, args = await self._parse(event)
            if not passed:
                return None
            self._invalid = True

        return await self._process(func, event, args)

    async def _process(
        self, func: AsyncCallable[..., bool | None], event: Event, args: AbstractParseArgs | None
    ) -> bool | None:
        event.spread = not self._block
        parse_args_ctx = ParseArgsCtx()
        if args is not None:
            args_token = parse_args_ctx.add(args)
        else:
            args_token = None

        try:
            if self._rule is not None:
                async with enter_session(self._rule):
                    return await func()
            return await func()
        finally:
            if args_token:
                parse_args_ctx.remove(args_token)

    async def _parse(self, event: Event) -> tuple[bool, AbstractParseArgs | None]:
        args: AbstractParseArgs | None = None
        if self.parser:
            event = cast(TextEvent, event)
            args = await self.parser.parse(event.text)
            if args:
                return (True, args)
            return (False, None)

        return (True, None)


def on_event(
    checker: Checker | None | SyncOrAsyncCallable[[Event], bool] = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    rule: Rule[Event] | None = None,
) -> FlowDecorator:
    """绑定任意事件的处理流装饰方法

    在前期的教程中，处理流装饰方法也称为绑定方法

    :param checker: 检查器
    :param priority: 优先级
    :param block: 是否阻断向低优先级传播
    :param temp: 是否临时使用（处理一次事件后停用）
    :param decos: 装饰器组
    :param rule: 会话规则
    :return: 处理流装饰器
    """
    return FlowDecorator(checker, None, None, priority, block, temp, decos, rule)


def on_text(
    checker: Checker | None | SyncOrAsyncCallable[[TextEvent], bool] = None,
    matcher: Matcher | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    """绑定文本事件的处理流装饰方法

    在前期的教程中，处理流装饰方法也称为绑定方法

    :param checker: 检查器
    :param matcher: 匹配器
    :param parser: 解析器
    :param priority: 优先级
    :param block: 是否阻断向低优先级传播
    :param temp: 是否临时使用（处理一次事件后停用）
    :param decos: 装饰器组
    :param legacy_session: 是否自动启用传统会话
    :return: 处理流装饰器
    """

    return FlowDecorator(
        checker_join(lambda e: isinstance(e, TextEvent), checker),  # type: ignore[arg-type]
        matcher,
        parser,
        priority,
        block,
        temp,
        decos,
        DefaultRule() if legacy_session else None,
    )


def on_command(
    cmd_start: str | Iterable[str],
    cmd_sep: str | Iterable[str],
    targets: str | Sequence[str],
    fmtters: Sequence[CmdArgFormatter | None] | None = None,
    checker: Checker | None | SyncOrAsyncCallable[[TextEvent], bool] = None,
    matcher: Matcher | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    """绑定文本事件的“命令解析”处理流装饰方法

    在前期的教程中，处理流装饰方法也称为绑定方法

    :param cmd_start: 命令起始符
    :param cmd_sep: 命令分隔符
    :param targets: 解析匹配的命令名
    :param fmtters: 命令参数格式化器
    :param checker: 检查器
    :param matcher: 匹配器
    :param priority: 优先级
    :param block: 是否阻断向低优先级传播
    :param temp: 是否临时使用（处理一次事件后停用）
    :param decos: 装饰器组
    :param legacy_session: 是否自动启用传统会话
    :return: 处理流装饰器
    """
    return FlowDecorator(
        checker_join(lambda e: isinstance(e, TextEvent), checker),  # type: ignore[arg-type]
        matcher,
        CmdParser(cmd_start=cmd_start, cmd_sep=cmd_sep, targets=targets, fmtters=fmtters),
        priority,
        block,
        temp,
        decos,
        DefaultRule() if legacy_session else None,
    )


def on_start_match(
    target: str | Sequence[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | SyncOrAsyncCallable[[TextEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    """绑定文本事件的“字符串起始匹配”处理流装饰方法

    在前期的教程中，处理流装饰方法也称为绑定方法

    :param target: 匹配的目标字符串
    :param logic_mode: 匹配的逻辑模式
    :param checker: 检查器
    :param parser: 解析器
    :param priority: 优先级
    :param block: 是否阻断向低优先级传播
    :param temp: 是否临时使用（处理一次事件后停用）
    :param decos: 装饰器组
    :param legacy_session: 是否自动启用传统会话
    :return: 处理流装饰器
    """
    return FlowDecorator(
        checker_join(lambda e: isinstance(e, TextEvent), checker),  # type: ignore[arg-type]
        StartMatcher(target, logic_mode),
        parser,
        priority,
        block,
        temp,
        decos,
        DefaultRule() if legacy_session else None,
    )


def on_contain_match(
    target: str | Sequence[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | SyncOrAsyncCallable[[TextEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    """绑定文本事件的“字符串包含匹配”处理流装饰方法

    在前期的教程中，处理流装饰方法也称为绑定方法

    :param target: 匹配的目标字符串
    :param logic_mode: 匹配的逻辑模式
    :param checker: 检查器
    :param parser: 解析器
    :param priority: 优先级
    :param block: 是否阻断向低优先级传播
    :param temp: 是否临时使用（处理一次事件后停用）
    :param decos: 装饰器组
    :param legacy_session: 是否自动启用传统会话
    :return: 处理流装饰器
    """
    return FlowDecorator(
        checker_join(lambda e: isinstance(e, TextEvent), checker),  # type: ignore[arg-type]
        ContainMatcher(target, logic_mode),
        parser,
        priority,
        block,
        temp,
        decos,
        DefaultRule() if legacy_session else None,
    )


def on_full_match(
    target: str | Sequence[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | SyncOrAsyncCallable[[TextEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    """绑定文本事件的“字符串完整匹配”处理流装饰方法

    在前期的教程中，处理流装饰方法也称为绑定方法

    :param target: 匹配的目标字符串
    :param logic_mode: 匹配的逻辑模式
    :param checker: 检查器
    :param parser: 解析器
    :param priority: 优先级
    :param block: 是否阻断向低优先级传播
    :param temp: 是否临时使用（处理一次事件后停用）
    :param decos: 装饰器组
    :param legacy_session: 是否自动启用传统会话
    :return: 处理流装饰器
    """
    return FlowDecorator(
        checker_join(lambda e: isinstance(e, TextEvent), checker),  # type: ignore[arg-type]
        FullMatcher(target, logic_mode),
        parser,
        priority,
        block,
        temp,
        decos,
        DefaultRule() if legacy_session else None,
    )


def on_end_match(
    target: str | Sequence[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | SyncOrAsyncCallable[[TextEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    """绑定文本事件的“字符串末尾匹配”处理流装饰方法

    在前期的教程中，处理流装饰方法也称为绑定方法

    :param target: 匹配的目标字符串
    :param logic_mode: 匹配的逻辑模式
    :param checker: 检查器
    :param parser: 解析器
    :param priority: 优先级
    :param block: 是否阻断向低优先级传播
    :param temp: 是否临时使用（处理一次事件后停用）
    :param decos: 装饰器组
    :param legacy_session: 是否自动启用传统会话
    :return: 处理流装饰器
    """
    return FlowDecorator(
        checker_join(lambda e: isinstance(e, TextEvent), checker),  # type: ignore[arg-type]
        EndMatcher(target, logic_mode),
        parser,
        priority,
        block,
        temp,
        decos,
        DefaultRule() if legacy_session else None,
    )


def on_regex_match(
    target: str,
    regex_flags: Any = 0,
    checker: Checker | SyncOrAsyncCallable[[TextEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    """绑定文本事件的“正则匹配”处理流装饰方法

    在前期的教程中，处理流装饰方法也称为绑定方法

    :param target: 匹配的目标字符串
    :param regex_flags: 正则匹配的 flags
    :param checker: 检查器
    :param parser: 解析器
    :param priority: 优先级
    :param block: 是否阻断向低优先级传播
    :param temp: 是否临时使用（处理一次事件后停用）
    :param decos: 装饰器组
    :param legacy_session: 是否自动启用传统会话
    :return: 处理流装饰器
    """
    return FlowDecorator(
        checker_join(lambda e: isinstance(e, TextEvent), checker),  # type: ignore[arg-type]
        RegexMatcher(target, regex_flags),
        parser,
        priority,
        block,
        temp,
        decos,
        DefaultRule() if legacy_session else None,
    )



from __future__ import annotations

import asyncio
import base64
import io
import time
import traceback
import warnings
from contextlib import asynccontextmanager
from datetime import datetime
from functools import wraps
from inspect import currentframe
from types import FrameType

from typing_extensions import Any, AsyncGenerator, Callable, Literal, cast

# 导入用作重导出（兼容过去的布局）
from .._lazy import singleton
from ..typ.base import P, T


def get_obj_name(
    obj: Any,
    otype: Literal["callable", "class", "object"] | str = "object",
    default: str = "<anonymous %s>",
) -> str:
    """获取一个对象的限定名称或名称，这适用于一些类型较宽的参数。

    无法获取有效名称时，产生一个 `default % otype` 字符串

    例如某处接受一个 `Callable` 类型的参数，对于一般函数来说，使用
    `__qualname__` 或 `__name__` 可获得名称，但某些可调用对象这些值可能为 `None`
    或不存在。使用此方法可保证一定返回字符串

    .. code:: python

        def _(a: Callable) -> None:
            valid_str: str = get_obj_name(a, otype="callable")

        def _(a: type) -> None:
            valid_str: str = get_obj_name(a, otype="class")

        def _(a: Any) -> None:
            valid_str: str = get_obj_name(a, otype="type of a, only for str concat")


    :param obj: 对象
    :param otype: 预期的对象类型
    :param default: 无法获取任何有效名称时的默认字符串
    :return: 对象名称或默认字符串
    """
    if hasattr(obj, "__qualname__"):
        return cast(str, obj.__qualname__)

    if hasattr(obj, "__name__"):
        return cast(str, obj.__name__)

    return default % otype


def deprecate_warn(msg: str, stacklevel: int = 2) -> None:
    from ..log.reflect import logger

    logger.warning(msg)
    warnings.simplefilter("always", DeprecationWarning)
    warnings.warn(msg, category=DeprecationWarning, stacklevel=stacklevel)
    warnings.simplefilter("default", DeprecationWarning)


def deprecated(msg: str) -> Callable[[Callable[P, T]], Callable[P, T]]:

    def deprecated_wrapper(func: Callable[P, T]) -> Callable[P, T]:

        @wraps(func)
        def deprecated_wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
            deprecate_warn(
                f"使用了弃用函数/方法 {func.__module__}.{func.__qualname__}: {msg}",
                stacklevel=3,
            )
            return func(*args, **kwargs)

        return deprecated_wrapped

    return deprecated_wrapper


class RWContext:
    """异步读写上下文

    提供异步安全的读写上下文。在读取时可以多读，同时读写互斥。

    使用方法：

    .. code:: python

       rwc = RWContext()
       # 读时使用此控制器的安全读上下文：
       async with rwc.read():
           ...
       # 写时使用此控制器的安全写上下文：
       async with rwc.write():
           ...
    """

    def __init__(self, read_limit: int | None = None) -> None:
        """初始化异步读写上下文

        :param read_limit: 读取的数量限制，为空则不限制
        """
        self.write_semaphore = asyncio.Semaphore(1)
        self.read_semaphore = asyncio.Semaphore(read_limit) if read_limit else None
        self.read_num = 0
        self.read_num_lock = asyncio.Lock()

    @asynccontextmanager
    async def read(self) -> AsyncGenerator[None, None]:
        """上下文管理器，展开一个关于该对象的安全异步读上下文"""
        if self.read_semaphore:
            await self.read_semaphore.acquire()

        async with self.read_num_lock:
            if self.read_num == 0:
                await self.write_semaphore.acquire()
            self.read_num += 1

        try:
            yield
        finally:
            async with self.read_num_lock:
                self.read_num -= 1
                if self.read_num == 0:
                    self.write_semaphore.release()
                if self.read_semaphore:
                    self.read_semaphore.release()

    @asynccontextmanager
    async def write(self) -> AsyncGenerator[None, None]:
        """上下文管理器，展开一个关于该对象的安全异步写上下文"""
        await self.write_semaphore.acquire()
        try:
            yield
        finally:
            self.write_semaphore.release()


class SnowFlakeIdWorker:
    def __init__(self, datacenter_id: int, worker_id: int, sequence: int = 0) -> None:
        self.max_worker_id = -1 ^ (-1 << 3)
        self.max_datacenter_id = -1 ^ (-1 << 5)
        self.worker_id_shift = 12
        self.datacenter_id_shift = 12 + 3
        self.timestamp_left_shift = 12 + 3 + 5
        self.sequence_mask = -1 ^ (-1 << 12)
        self.startepoch = int(datetime(2022, 12, 11, 12, 8, 45).timestamp() * 1000)

        if worker_id > self.max_worker_id or worker_id < 0:
            raise ValueError("worker_id 值越界")
        if datacenter_id > self.max_datacenter_id or datacenter_id < 0:
            raise ValueError("datacenter_id 值越界")
        self.worker_id = worker_id
        self.datacenter_id = datacenter_id
        self.sequence = sequence
        self.last_timestamp = -1

    def _gen_timestamp(self) -> int:
        return int(time.time() * 1000)

    def get_id(self) -> int:
        timestamp = self._gen_timestamp()

        if timestamp < self.last_timestamp:
            raise ValueError(f"时钟回拨，{self.last_timestamp} 前拒绝 id 生成请求")
        if timestamp == self.last_timestamp:
            self.sequence = (self.sequence + 1) & self.sequence_mask
            if self.sequence == 0:
                timestamp = self._until_next_millis(self.last_timestamp)
        else:
            self.sequence = 0
        self.last_timestamp = timestamp
        new_id = (
            ((timestamp - self.startepoch) << self.timestamp_left_shift)
            | (self.datacenter_id << self.datacenter_id_shift)
            | (self.worker_id << self.worker_id_shift)
            | self.sequence
        )
        return new_id

    def get_b64_id(self, trim_pad: bool = True) -> str:
        id = base64.urlsafe_b64encode(self.get_id().to_bytes(8, byteorder="little")).decode()
        if trim_pad:
            id = id.rstrip("=")
        return id

    def _until_next_millis(self, last_time: int) -> int:
        timestamp = self._gen_timestamp()
        while timestamp <= last_time:
            timestamp = self._gen_timestamp()
        return timestamp


_DEFAULT_ID_WORKER = SnowFlakeIdWorker(1, 1, 0)


def get_id() -> str:
    """从 melobot 内部 id 获取器获得一个 id 值，不保证线程安全。算法使用雪花算法

    :return: id 值
    """
    return _DEFAULT_ID_WORKER.get_b64_id()


def find_caller_stack(
    stack_info: bool = False,
    stacklevel: int = 1,
    inner_frame_filter: Callable[[FrameType], bool] | None = None,
) -> tuple[str, str, int, str, str | None]:
    f = currentframe()
    if f is None:
        return "<unknown module>", "<unknown file>", 0, "<unknown function>", "<unknown stackinfo>"

    while stacklevel > 0:
        next_f = f.f_back
        if next_f is None:
            break
        f = next_f
        if inner_frame_filter is None or not inner_frame_filter(f):
            stacklevel -= 1
    co = f.f_code
    sinfo = None

    if stack_info:
        with io.StringIO() as sio:
            sio.write("Stack (most recent call last):\n")
            traceback.print_stack(f, file=sio)
            sinfo = sio.getvalue()
            if sinfo[-1] == "\n":
                sinfo = sinfo[:-1]

    if not isinstance(f.f_lineno, int):
        raise ValueError(f"尝试解析调用栈时，获取的行号不是整数，值类型是：{type(f.f_lineno)}")
    return f.f_globals["__name__"], co.co_filename, f.f_lineno, co.co_name, sinfo



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from pathlib import Path

from typing_extensions import Callable, Iterable, final, overload

from ..exceptions import PluginLoadError
from ..handle.base import Flow
from ..mixin import HookMixin
from ..typ.base import P, T
from .ipc import AsyncShare, SyncShare


class PluginLifeSpan(Enum):
    """插件生命周期的枚举"""

    INITED = "i"


@dataclass(frozen=True)
class PluginInfo:
    """插件信息类，用于添加描述信息"""

    desc: str = ""
    docs: Path | None = None
    keywords: tuple[str] | None = None
    url: str = ""
    author: str = ""


class PluginPlanner(HookMixin[PluginLifeSpan]):
    """插件管理器类

    用于声明一个插件，并为插件添加功能
    """

    def __init__(
        self,
        version: str,
        flows: Iterable[Flow] | None = None,
        shares: Iterable[SyncShare | AsyncShare] | None = None,
        funcs: Iterable[Callable] | None = None,
        auto_import: list[str] | bool = False,
        *,
        info: PluginInfo | None = None,
    ) -> None:
        """插件管理器初始化

        :param version: 版本号
        :param flows: 事件流。可以先指定为空，后续使用 :meth:`use` 绑定
        :param shares: 共享对象（需要在本插件内定义）。可以先指定空，后续用 :meth:`use` 绑定
        :param funcs: 导出函数（需要在本插件内定义，提供方法是未定义行为）。可以先指定空，后续用 :meth:`use` 绑定
        :param auto_import:
            需要自动导入的模块的路径列表（相对路径以插件目录为基准），该参数对动态插件无效。
            如果为 `True` 导入插件目录下所有模块，此时只会导入 `.py` 模块。
            如果你需要导入 `.{pyc,pyd,so,...}` 等其他可加载模块，请自行提供列表。自行提供列表时的一些提示：

            不要包含目录路径，这永远没有效果

            建议使用 :func:`glob.glob` 或 :meth:`pathlib.Path.glob` 方法获取路径，而不是手动拼接或查找

            一个模块在加载时，其向上到插件目录的所有父目录的 `__init__.{pyc,pyd,so,py...}` 都会被自动加载，
            此时不需要手动提供 `__init__.{pyc,pyd,so,py...}` 文件。加载时的扩展名优先级请查看 :data:`~.MODULE_EXTS`
            （优先级从高到低，且与操作系统平台有关）

            如果一个目录中只有 `__init__.{pyc,pyd,so,py...}` 文件，此时只能手动提供

        :param info: 插件信息
        """
        super().__init__(hook_type=PluginLifeSpan)
        self.version = version
        self.init_flows = [] if flows is None else list(flows)
        self.shares = set[SyncShare | AsyncShare]() if shares is None else set(shares)
        self.funcs = set[Callable]() if funcs is None else set(funcs)
        self.info = PluginInfo() if info is None else info
        self.auto_import = auto_import

        self._pname: str = ""
        self._built: bool = False
        self._plugin: Plugin

    @overload
    def use(self, obj: Flow) -> Flow: ...
    @overload
    def use(self, obj: SyncShare[T]) -> SyncShare[T]: ...  # type: ignore[overload-overlap]
    @overload
    def use(self, obj: AsyncShare[T]) -> AsyncShare[T]: ...  # type: ignore[overload-overlap]
    @overload
    def use(self, obj: Callable[P, T]) -> Callable[P, T]: ...

    @final
    def use(self, obj: T) -> T:
        """装饰器

        绑定一个组件（流，共享对象，导出函数），标记插件创建后使用该组件。

        :param obj: 可用的组件
        :return: 被绑定的组件本身
        """
        if isinstance(obj, Flow):
            self.init_flows.append(obj)
        elif isinstance(obj, (SyncShare, AsyncShare)):
            self.shares.add(obj)
        elif callable(obj):
            self.funcs.add(obj)
        else:
            raise PluginLoadError(f"插件无法使用 {type(obj)} 类型的对象")
        return obj


class Plugin:
    def __init__(self, planner: PluginPlanner) -> None:
        self.planner = planner
        self.name = planner._pname
        self.hook_bus = planner._hook_bus

        self.shares = planner.shares
        self.funcs = planner.funcs
        self.init_flows = planner.init_flows



from asyncio import run


async def get_sum(r_bound: int) -> int:
    return sum(range(1, r_bound + 1))


def sync_get_sum(r_bound: int) -> int:
    return sum(range(1, r_bound + 1))


def main():
    r_bound = 10000
    result = run(get_sum(r_bound))
    print(f"\nSum of numbers from 1 to {r_bound} is: {result}")



"""Cross platform Sphinx doc build, called by uv."""

import os
import sys
from pathlib import Path

os.chdir(Path(__file__).parent.absolute().resolve(strict=True).as_posix())


def main():
    if sys.platform != "win32":
        os.system("make html")
    else:
        os.system("make.bat html")


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass
from itertools import tee

from typing_extensions import Any, Generic, Hashable, Iterable, Iterator, Sequence, TypeIs, TypeVar

T = TypeVar("T", bound=Hashable)


@dataclass
class NodeInfo(Generic[T]):
    nexts: list[T]
    in_deg: int
    out_deg: int

    def copy(self) -> NodeInfo[T]:
        return NodeInfo(self.nexts, self.in_deg, self.out_deg)


def _is_iterable(obj: Any) -> TypeIs[Iterable]:
    # 这个操作实际上不是绝对安全，获取 iter 可能是有副作用的
    # 但是这种可能性非常非常小，所以不予考虑
    # 如果确实有相关反馈，那么再调整
    try:
        iter(obj)  # type: ignore[call-overload]
    except TypeError:
        return False
    else:
        return True


class DAGMapping(Generic[T]):
    def __init__(self, name: str, *edges: Iterable[Iterable[T] | T]) -> None:
        self.name = name
        self.map: dict[T, NodeInfo[T]] = {}
        self._verified = False

        _edges = tuple(
            tuple((elem,) if not _is_iterable(elem) else elem for elem in emap) for emap in edges
        )
        edge_pairs = self._get_edges(_edges)
        for n1, n2 in edge_pairs:
            self.add(n1, n2)
        self.verify()

    def __len__(self) -> int:
        return len(self.map)

    def __getitem__(self, key: T) -> NodeInfo[T]:
        return self.map[key]

    def __contains__(self, key: T) -> bool:
        return key in self.map

    def __iter__(self) -> Iterator[tuple[T, NodeInfo[T]]]:
        return iter(self.map.items())

    @property
    def starts(self) -> tuple[T, ...]:
        return tuple(n for n, info in self.map.items() if info.in_deg == 0)

    @property
    def ends(self) -> tuple[T, ...]:
        return tuple(n for n, info in self.map.items() if info.out_deg == 0)

    def _get_edges(self, edge_maps: Sequence[Sequence[Iterable[T]]]) -> list[tuple[T, T]]:
        edges: list[tuple[T, T]] = []

        for emap in edge_maps:
            iter1, iter2 = tee(emap, 2)
            try:
                next(iter2)
            except StopIteration:
                continue
            if len(emap) == 1:
                for n in emap[0]:
                    self.add(n, None)
                continue

            for from_seq, to_seq in zip(iter1, iter2):
                for n1 in from_seq:
                    for n2 in to_seq:
                        if (n1, n2) not in edges:
                            edges.append((n1, n2))
        return edges

    def add(self, _from: T, to: T | None) -> None:
        from_info = self.map.setdefault(_from, NodeInfo([], 0, 0))
        if to is not None:
            to_info = self.map.setdefault(to, NodeInfo([], 0, 0))
            to_info.in_deg += 1
            from_info.out_deg += 1
            from_info.nexts.append(to)
        self._verified = False

    def verify(self) -> None:
        if self._verified:
            return

        graph = {n: info.copy() for n, info in self.map.items()}
        type_set = {type(n) for n in self.map}
        if len(type_set) > 1:
            type_dic: dict[type, list[Any]] = {}
            for n in self.map:
                type_dic.setdefault(type(n), []).append(n)
            raise TypeError(
                f"名为 {self.name} 的图结构中，结点类型不一致。"
                f"类型有：{type_set}。"
                f"每种类型的结点如下：{dict(sorted(type_dic.items(), key=lambda x: len(x[1])))}"
            )

        while len(graph):
            for n, info in graph.items():
                nexts, in_deg = info.nexts, info.in_deg
                if in_deg == 0:
                    graph.pop(n)
                    for next_n in nexts:
                        graph[next_n].in_deg -= 1
                    break
            else:
                raise ValueError(f"名为 {self.name} 的图结构中存在环路")
        self._verified = True

    def link(self, graph: DAGMapping[T], name: str) -> DAGMapping[T]:
        _froms = self.ends
        _tos = graph.starts
        new_edges = tuple((n1, n2) for n1 in _froms for n2 in _tos)
        new_graph = DAGMapping(name, *new_edges)

        for n1, info in (self.map | graph.map).items():
            if not len(info.nexts):
                new_graph.add(n1, None)
                continue
            for n2 in info.nexts:
                new_graph.add(n1, n2)

        new_graph.verify()
        return new_graph



import warnings
from functools import wraps

from typing_extensions import Any, Callable, Iterable, ParamSpec, TypeVar, cast, overload

T = TypeVar("T", default=Any)
P = ParamSpec("P", default=Any)

# 以下代码，由 websockets.imports 模块源代码修改而来
# 原始版权 © Aymeric Augustin and contributors
# 原始许可：https://github.com/python-websockets/websockets/blob/main/LICENSE


def import_name(name: str, source: str, namespace: dict[str, Any]) -> Any:
    level = 0
    while source[level] == ".":
        level += 1
        if level >= len(source):
            raise ImportError("importing from parent isn't supported")
    module = __import__(source[level:], namespace, None, [name], level)
    return getattr(module, name)


def lazy_import(
    mod_globals: dict[str, Any],
    map: dict[str, tuple[str, ...]],
    deprecations: dict[str, tuple[tuple[str, str], ...]],
) -> None:
    # 不公开给用户，仅 melobot 内部使用
    mapping: dict[str, str] = {}
    for location, names in map.items():
        for name in names:
            mapping[name] = location

    deprecated_dic: dict[str, tuple[str, str]] = {}
    for location, pairs in deprecations.items():
        for name, ver in pairs:
            deprecated_dic[name] = (location, ver)

    mod_globals_set = set(mod_globals)
    mapping_set = set(mapping)
    deprecated_set = set(deprecated_dic)

    if mod_globals_set & mapping_set:
        raise ValueError("原始模块与延迟加载项有冲突的命名")
    if mod_globals_set & deprecated_set:
        raise ValueError("原始模块与弃用项有冲突的命名")
    if mapping_set & deprecated_set:
        raise ValueError("延迟加载项与弃用项有冲突的命名")

    mod_name = mod_globals["__name__"]

    def __getattr__(name: str) -> Any:
        try:
            location = mapping[name]
        except KeyError:
            pass
        else:
            return import_name(name, location, mod_globals)

        try:
            location, ver = deprecated_dic[name]
        except KeyError:
            pass
        else:
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                f"{mod_name}.{name} 现以弃用，将于 {ver} 版本移除",
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return import_name(name, location, mod_globals)

        raise AttributeError(f"module {mod_name!r} has no attribute {name!r}")

    def __dir__() -> Iterable[str]:
        return sorted(mod_globals_set | mapping_set | deprecated_set)

    mod_globals["__getattr__"] = __getattr__
    mod_globals["__dir__"] = __dir__


# --------------------------------------------------------------------------


class LazyLoader:
    def __init__(
        self,
        namespace: dict[str, Any],
        location: str,
        item: str | None = None,
        alias: str | None = None,
    ) -> None:
        self.item = item
        self.alias = alias
        self.location = location
        if self.location.startswith("."):
            raise ValueError(f"延迟加载不支持 {location!r} 中的相对引用语义")
        self.loc_parts = location.split(".")
        self.namespace = namespace
        self._value: Any

        if self.alias is not None:
            self.namespace[self.alias] = self
        elif self.item is None:
            self.namespace[self.loc_parts[0]] = self
        else:
            self.namespace[self.item] = self

    def __repr__(self) -> str:
        statement = ""
        if self.item is None:
            statement += f"import {self.location}"
        else:
            statement += f"from {self.location} import {self.item}"
        if self.alias is not None:
            statement += f" as {self.alias}"
        return f"{self.__class__.__name__}(equation={statement!r})"

    def _load(self) -> None:
        module = __import__(
            self.location, self.namespace, None, [self.item] if self.item is not None else ()
        )

        if self.item is None:
            if self.alias is None:
                self.namespace[self.loc_parts[0]] = self._value = module
            else:
                if len(self.loc_parts) > 1:
                    node = module
                    idx = 1
                    while idx < len(self.loc_parts):
                        node = getattr(node, self.loc_parts[idx])
                        idx += 1
                    module = node
                self.namespace[self.alias] = self._value = module
            return

        obj = getattr(module, self.item)
        if self.alias is None:
            self.namespace[self.item] = self._value = obj
        else:
            self.namespace[self.alias] = self._value = obj

    def __getattr__(self, name: str) -> Any:
        self._load()
        return getattr(self._value, name)


def lazy_load(
    namespace: dict[str, Any], location: str, item: str | None = None, alias: str | None = None
) -> str:
    """惰性加载模块的方法

    指定的模块或者对象，在发生第一次属性访问时触发加载，并同步修改 `namespace` 字典中对应键值。

    在完成加载后与常规的 import 语句效果无异，但需要略微注意未加载时的一些行为。
    例如未加载时 `print` 对象，行为可能就不一致。

    .. code:: python

        # 典型的使用方法：
        if TYPE_CHECKING:
            import matplotlib as mpl
            import matplotlib.font_manager as fm
            import PIL
            from matplotlib import pyplot as plt
        else:
            _g = globals()
            lazy_load(_g, "matplotlib", alias="mpl")
            lazy_load(_g, "matplotlib.font_manager", alias="fm")
            lazy_load(_g, "PIL")
            lazy_load(_g, "matplotlib", item="pyplot", alias="plt")

    假设有导入语句：

    `import xxx`

    `import xxx as zzz`

    `from xxx import yyy`

    `from xxx import yyy as zzz`

    :param namespace: 当前模块的 globals() 字典
    :param location: import 语句中 `xxx` 的部分
    :param item: import 语句中 `yyy` 的部分，无则为空
    :param alias: import 语句中 `zzz` 的部分，无则为空
    :return: 惰性加载对象的 repr 字符串，用于调试
    """
    return repr(LazyLoader(namespace, location, item, alias))


_SINGLETON_OBJ_MAP: dict[Any, Any] = {}
_SINGLETON_FACTORY_MAP: dict[Any, Any] = {}


@overload
def singleton(cls: type[T]) -> type[T]: ...
@overload
def singleton(cls: Callable[P, T]) -> Callable[P, T]: ...


def singleton(cls: type[T] | Callable[P, T]) -> type[T] | Callable[P, T]:
    """单例装饰器

    :param cls: 需要被单例化的可调用对象
    :return: 需要被单例化的可调用对象
    """

    @wraps(cls)
    def singleton_wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        if cls not in _SINGLETON_OBJ_MAP:
            obj = _SINGLETON_OBJ_MAP[cls] = cls(*args, **kwargs)
            _SINGLETON_FACTORY_MAP[obj] = cls
        return cast(T, _SINGLETON_OBJ_MAP[cls])

    return singleton_wrapped


def singleton_clear(obj: Any) -> None:
    """清除已经缓存的单例

    :param obj: 单例化得到的对象
    """
    if obj not in _SINGLETON_FACTORY_MAP:
        raise ValueError(
            f"{obj} 不在单例装饰器的记录中。"
            "它可能不是单例对象，或产生此单例的类或可调用对象已清空单例缓存"
        )
    callable_or_cls = _SINGLETON_FACTORY_MAP.pop(obj)
    _SINGLETON_OBJ_MAP.pop(callable_or_cls)



from melobot.protocols import ProtocolStack

from .. import __version__
from .adapter import *  # noqa: F403
from .const import PROTOCOL_IDENTIFIER, PROTOCOL_NAME, PROTOCOL_SUPPORT_AUTHOR, PROTOCOL_VERSION
from .handle import on_at_qq, on_event, on_message, on_meta, on_notice, on_request
from .io import *  # noqa: F403
from .utils import *  # noqa: F403


class OneBotV11Protocol(ProtocolStack):
    def __init__(self, *srcs: BaseSource) -> None:
        super().__init__()
        self.adapter = Adapter()
        self.inputs = set()
        self.outputs = set()

        for src in srcs:
            if not isinstance(src, BaseSource):
                raise TypeError(f"不支持的 OneBot v11 源类型: {type(src)}")
            if isinstance(src, BaseInSource):
                self.inputs.add(src)
            if isinstance(src, BaseOutSource):
                self.outputs.add(src)



from __future__ import annotations

import re
from abc import abstractmethod
from functools import reduce

from typing_extensions import Any, Sequence, assert_never

from ...typ._enum import LogicMode
from ...typ.cls import BetterABC


class Matcher(BetterABC):
    """匹配器基类"""

    def __init__(self) -> None:
        super().__init__()

    def __and__(self, other: Matcher) -> WrappedMatcher:
        if not isinstance(other, Matcher):
            return NotImplemented
        return WrappedMatcher(LogicMode.AND, self, other)

    def __or__(self, other: Matcher) -> WrappedMatcher:
        if not isinstance(other, Matcher):
            return NotImplemented
        return WrappedMatcher(LogicMode.OR, self, other)

    def __invert__(self) -> WrappedMatcher:
        return WrappedMatcher(LogicMode.NOT, self)

    def __xor__(self, other: Matcher) -> WrappedMatcher:
        if not isinstance(other, Matcher):
            return NotImplemented
        return WrappedMatcher(LogicMode.XOR, self, other)

    @abstractmethod
    async def match(self, text: str) -> bool:
        """匹配器匹配方法

        任何匹配器应该实现此抽象方法。

        :param text: 消息事件的文本内容
        :return: 是否匹配
        """
        raise NotImplementedError


class WrappedMatcher(Matcher):
    """合并匹配器

    在两个 :class:`Matcher` 对象间使用 | & ^ ~ 运算符即可返回合并匹配器
    """

    def __init__(
        self,
        mode: LogicMode,
        matcher1: Matcher,
        matcher2: Matcher | None = None,
    ) -> None:
        """初始化一个合并匹配器

        :param mode: 合并匹配的逻辑模式
        :param matcher1: 匹配器1
        :param matcher2: 匹配器2
        """
        super().__init__()
        self.mode = mode
        self.m1, self.m2 = matcher1, matcher2

    async def match(self, text: str) -> bool:
        match self.mode:
            case LogicMode.AND:
                status = await self.m1.match(text) and await self.m2.match(text)  # type: ignore[union-attr]
            case LogicMode.OR:
                status = await self.m1.match(text) or await self.m2.match(text)  # type: ignore[union-attr]
            case LogicMode.NOT:
                status = not await self.m1.match(text)
            case LogicMode.XOR:
                status = await self.m1.match(text) ^ await self.m2.match(text)  # type: ignore[union-attr]
            case _:
                assert_never(f"无效的逻辑模式 {self.mode}")
        return status


class StartMatcher(Matcher):
    """字符串起始匹配器"""

    def __init__(self, target: str | Sequence[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串起始匹配器

        `target` 为字符串时，只进行一次起始匹配，即判断是否匹配成功。
        `target` 为字符串序列时，所有字符串都进行起始匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.mode = mode
        self.target: set[str] | str
        if not isinstance(target, str):
            self.target = set(target)
        else:
            self.target = target

    async def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return text.startswith(self.target)
        res_seq = tuple(text.startswith(s) for s in self.target)
        return reduce(self.mode.get_operator(), res_seq)


class ContainMatcher(Matcher):
    """字符串包含匹配器"""

    def __init__(self, target: str | Sequence[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串包含匹配器

        `target` 为字符串时，只进行一次包含匹配，即判断是否匹配成功。
        `target` 为字符串序列时，所有字符串都进行包含匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.mode = mode
        self.target: set[str] | str
        if not isinstance(target, str):
            self.target = set(target)
        else:
            self.target = target

    async def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return self.target in text
        res_seq = tuple(s in text for s in self.target)
        return reduce(self.mode.get_operator(), res_seq)


class EndMatcher(Matcher):
    """字符串结尾匹配器"""

    def __init__(self, target: str | Sequence[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串结尾匹配器

        `target` 为字符串时，只进行一次结尾匹配，即判断是否匹配成功。
        `target` 为字符串序列时，所有字符串都进行结尾匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.mode = mode
        self.target: set[str] | str
        if not isinstance(target, str):
            self.target = set(target)
        else:
            self.target = target

    async def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return text.endswith(self.target)
        res_seq = tuple(text.endswith(s) for s in self.target)
        return reduce(self.mode.get_operator(), res_seq)


class FullMatcher(Matcher):
    """字符串全匹配器"""

    def __init__(self, target: str | Sequence[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串全匹配器

        `target` 为字符串时，只进行一次全匹配，即判断是否匹配成功。
        `target` 为字符串序列时，所有字符串都进行全匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.mode = mode
        self.target: set[str] | str
        if not isinstance(target, str):
            self.target = set(target)
        else:
            self.target = target

    async def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return text == self.target
        res_seq = [text == s for s in self.target]
        return reduce(self.mode.get_operator(), res_seq)


class RegexMatcher(Matcher):
    """字符串正则匹配器"""

    def __init__(self, regex_pattern: str, regex_flags: Any = 0) -> None:
        """初始化一个字符串正则匹配器

        :param regex_pattern: 正则 pattern
        :param regex_flags: 正则 flag，默认不使用
        """
        super().__init__()
        self.pattern = regex_pattern
        self.flag = regex_flags

    async def match(self, text: str) -> bool:
        return len(re.findall(self.pattern, text, self.flag)) > 0



"""
此模块提供在 melobot 用例中对多进程的支持。
主要实现 spawn 模式子进程的自定义入口，以及避免默认序列化方式导致的级联加载。
基于对 spawn 模式子进程入口点的劫持，以及构建自定义可序列化对象。
只要不使用本模块的进程创建接口，就自动回退到 multiprocessing 原始逻辑
"""

# TODO: 由于存在侵入性设计，因此每个 py minor 版本都应该测试验证
import multiprocessing.spawn as spawn_mod
import multiprocessing.util as util_mod
import sys
from functools import wraps
from multiprocessing import current_process
from pathlib import Path

from typing_extensions import Any

from ._imp import ALL_EXTS

_PNAME_PREFIX = "MeloBot_MP"
MP_MODULE_NAME = "__mp_main__"
ROOT_MODULE_DIR = Path(__file__).parent.parent.resolve().as_posix()


def in_main_process() -> bool:
    """判断当前进程是否为主进程"""
    return current_process().name == "MainProcess"


def _wrapped_get_preparation_data(name: str) -> dict:
    data = _original_get_preparation_data(name)
    if SpawnProcess.owned(name):
        data["sys_path"].insert(0, _P_STATUS[name]["dir"])
        data["sys_argv"] = _P_STATUS[name]["argv"]
        data["dir"] = data["orig_dir"] = _P_STATUS[name]["dir"]

        sentinel = object()
        mod_name = data.get("init_main_from_name", sentinel)
        if mod_name is not sentinel and mod_name not in ("__main__", MP_MODULE_NAME):
            raise RuntimeError(
                f"子进程中 __main__ 模块从名称 {mod_name!r} 加载，这种情况下无法安全生成子进程"
            )
        data.pop("init_main_from_name", None)
        data["init_main_from_path"] = _P_STATUS[name]["entry"]
    return data


def _wrapped_get_command_line(**kwargs: Any) -> Any:
    if getattr(sys, "frozen", False):
        return [sys.executable, "--multiprocessing-fork"] + [
            "%s=%r" % item for item in kwargs.items()
        ]
    else:
        prog = (
            f"import sys; sys.path.insert(0, {ROOT_MODULE_DIR!r}); import melobot.mp; "
            "from multiprocessing.spawn import spawn_main; spawn_main(%s)"
        )
        prog %= ", ".join("%s=%r" % item for item in kwargs.items())
        opts = util_mod._args_from_interpreter_flags()  # type: ignore[attr-defined]
        return [spawn_mod.get_executable()] + opts + ["-c", prog, "--multiprocessing-fork"]


def _wrapped_prepare(data: Any) -> None:
    ret = _original_prepare(data)
    if SpawnProcess.owned(data["name"]):
        import signal

        # 默认重置常见的信号处理，子进程由父进程全权管理
        signals_to_ignore: list[signal.Signals] = [signal.SIGINT, signal.SIGTERM]
        if sys.platform == "win32":
            signals_to_ignore.append(signal.SIGBREAK)
        for sig in signals_to_ignore:
            signal.signal(sig, signal.SIG_IGN)
    return ret


_original_get_preparation_data = spawn_mod.get_preparation_data
spawn_mod.get_preparation_data = wraps(_original_get_preparation_data)(
    _wrapped_get_preparation_data
)

_original_get_command_line = spawn_mod.get_command_line
spawn_mod.get_command_line = wraps(_original_get_command_line)(_wrapped_get_command_line)

_original_prepare = spawn_mod.prepare
spawn_mod.prepare = wraps(_original_prepare)(_wrapped_prepare)


import pickle
from concurrent.futures import ProcessPoolExecutor as _ProcessPoolExecutor
from functools import partial
from multiprocessing import get_context
from multiprocessing.context import SpawnContext as _SpawnContext
from multiprocessing.pool import Pool
from os import PathLike
from os.path import normpath
from pathlib import Path
from threading import RLock
from types import FunctionType, MethodType, ModuleType

from typing_extensions import Callable, Iterable, Mapping, TypeAlias, TypedDict, cast


class _ProcessStatus(TypedDict):
    name: str
    entry: str
    argv: list[str]
    dir: str


_P_STATUS: dict[str, _ProcessStatus] = {}


class SpawnProcess(get_context("spawn").Process):  # type: ignore[name-defined,misc]
    """melobot 进程类（使用 spawn 模式生成子进程）"""

    def __init__(
        self,
        entry: str | PathLike[str] | Path,
        argv: list[str] | None = None,
        target: Callable[..., object] | None = None,
        name: str | None = None,
        args: Iterable[Any] | None = None,
        kwargs: Mapping[str, Any] | None = None,
        *,
        daemon: bool | None = None,
    ) -> None:
        """初始化一个子进程对象

        :param entry: 子进程的入口模块（必须是文件）
        :param argv: 子进程的 `argv`，为空时使用默认设置
        :param target: 子进程的目标函数
        :param name: 子进程的标识名称，注意真实的进程名称与此参数不相等，只是包含
        :param args: 目标函数的参数
        :param kwargs: 目标函数的参数
        :param daemon: 是否是守护进程
        """
        if not in_main_process():
            raise RuntimeError(
                "不应该在 melobot 管理的子进程中继续创建子进程。"
                "出现此异常可能是因为初始化参数“入口路径”设置错误，导致创建进程的代码在子进程中再次执行"
            )

        super().__init__(
            None,
            target,
            name,
            () if args is None else args,
            {} if kwargs is None else kwargs,
            daemon=daemon,
        )
        order = self.name.split("-")[-1]
        self.init_name = name
        # 重设 name 属性，用于后续在 hack 中区分进程
        self.main_part = f"{id(self):x}" if self.init_name is None else self.init_name
        self.name: str = f"{_PNAME_PREFIX}_{self.main_part}-{order}"

        try:
            entry_file = Path(entry).resolve(True)
        except FileNotFoundError as e:
            raise FileNotFoundError(f"子进程 {self.main_part} 的入口不存在: {entry!r}") from e

        if not entry_file.is_file() or not entry_file.as_posix().endswith(ALL_EXTS):
            raise ValueError(f"子进程 {self.main_part} 的入口不是可加载的文件: {entry_file!r}")

        # 再使用一次 normpath，保证与原始实现一样的兼容性
        entry_norm_path = normpath(entry_file.as_posix())
        _P_STATUS[self.name] = {
            "name": self.name,
            "entry": entry_norm_path,
            "argv": argv if argv is not None else [entry_norm_path],
            "dir": normpath(entry_file.parent.as_posix()),
        }

    @staticmethod
    def owned(name: str) -> bool:
        return "melobot" in name.lower()


def in_melobot_sub_process() -> bool:
    """判断当前进程是否为 melobot 管理的子进程"""
    return SpawnProcess.owned(current_process().name)


class _BanOriginalProcess:
    def __get__(self, *_: Any, **__: Any) -> None:
        raise AttributeError(f"内部尝试引用原始 Process 对象，请报告关于 {__name__} 模块的 bug")


class SpawnContext(_SpawnContext):
    Process = cast(Any, _BanOriginalProcess())

    def __init__(self, entry: str | PathLike[str] | Path, argv: list[str] | None = None) -> None:
        super().__init__()
        self.process_entry = entry
        self.process_argv = argv

    def __getattr__(self, name: str) -> Any:
        if name == "Process":
            return partial(SpawnProcess, self.process_entry, self.process_argv)
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")


class SpawnProcessPool(Pool):
    """melobot 进程池类（使用 spawn 模式生成子进程）"""

    def __init__(
        self,
        entry: str | PathLike[str] | Path,
        argv: list[str] | None = None,
        processes: int | None = None,
        initializer: Callable[..., object] | None = None,
        initargs: Iterable[Any] | None = None,
        maxtasksperchild: int | None = None,
    ) -> None:
        """初始化一个进程池对象

        :param entry: 所有子进程的入口模块（必须是文件）
        :param argv: 所有子进程的 `argv`，为空时使用默认设置
        :param processes: 进程数
        :param initializer: 初始化函数
        :param initargs: 初始化函数的参数
        :param maxtasksperchild: 每个子进程执行的任务数，达到此数时销毁并生成新进程
        """
        if initargs is None:
            init_args = ()
        super().__init__(
            processes, initializer, init_args, maxtasksperchild, SpawnContext(entry, argv)
        )


class SpawnProcessPoolExecutor(_ProcessPoolExecutor):
    """melobot 进程池执行器类（使用 spawn 模式生成子进程）"""

    def __init__(
        self,
        entry: str | PathLike[str] | Path,
        argv: list[str] | None = None,
        max_workers: int | None = None,
        initializer: Callable[..., object] | None = None,
        initargs: Iterable[Any] | None = None,
    ) -> None:
        """初始化一个进程池执行器对象

        :param entry: 所有子进程的入口模块（必须是文件）
        :param argv: 所有子进程的 `argv`，为空时使用默认设置
        :param max_workers: worker（进程）的数量
        :param initializer: 初始化函数
        :param initargs: 初始化函数的参数
        """
        if initargs is None:
            init_args = ()
        super().__init__(max_workers, SpawnContext(entry, argv), initializer, init_args)


Process: TypeAlias = SpawnProcess
Context: TypeAlias = SpawnContext
ProcessPool: TypeAlias = SpawnProcessPool
ProcessPoolExecutor: TypeAlias = SpawnProcessPoolExecutor


_EMPTY = object()
_DUMMY_CLS = type("_DUMMY_CLS", (), {})
# 注意这个递归锁在序列化和反序列化时，不是同一个（在不同进程中）
_PICKLE_RLOCK = RLock()


class PBox:
    def __init__(
        self,
        value: Any = _EMPTY,
        name: str | None = None,
        module: str | None = None,
        entry: str | PathLike[str] | Path | None = None,
    ) -> None:
        """pickle 包装器

        更改或指定需要 pickle 的对象的来源

        :param value: 和 `name` 二选一，指定需要 pickle 的对象，默认值为一个哨兵对象（表示无值）
        :param name: 和 `value` 二选一，指定需要 pickle 的对象的名称
        :param module: pickle 的对象的来源模块名，为空时映射到进程的入口模块
        :param entry: pickle 的对象的来源模块的路径，如果为空则只依赖于模块名进行加载
        """
        if value is _EMPTY and name is None:
            raise ValueError("值参数和名称参数不能同时为空")
        if value is not _EMPTY and name is not None:
            raise ValueError("值参数和名称参数不能同时存在")
        if isinstance(value, MethodType):
            raise ValueError("类或实例的方法不支持 pickle，请尝试 pickle 整个类或实例而不是方法")

        # 序列化后的 bytes
        self.value: bytes
        # 是否有值，如果无值则反序列化时不使用 value 属性，而是直接从模块中提取
        self.has_value: bool
        # 反序列化前，预先加载的模块名
        self.module = cast(str, MP_MODULE_NAME if module in ("", None) else module)

        # 反序列化时，预先加载的模块的文件路径，为空时只依赖模块名加载模块
        self.entry: str | None
        if entry is not None:
            try:
                abs_entry_path = Path(entry).resolve(True)
            except FileNotFoundError as e:
                raise FileNotFoundError(f"入口路径 {entry!r} 不存在") from e
            self.entry = abs_entry_path.as_posix()

            if self.module != MP_MODULE_NAME:
                mod_parts = tuple(self.module.split("."))
                if "" in mod_parts:
                    raise ValueError(f"模块名 {module!r} 有误或存在相对导入语义")

                path_parts = abs_entry_path.parts[:-1] + (abs_entry_path.stem,)
                if path_parts[-len(mod_parts) :] != mod_parts:
                    raise ValueError(
                        "模块名 split('.') 后的序列，不是路径绝对化并去除扩展名后 split('/') 得到序列的尾子序列"
                    )
        else:
            self.entry = None

        # 这些属性在序列化时会被排除在外
        self._serial_args = (value, name)
        self._orig_value = value

    def _serialize(self, value: Any, name: str | None) -> None:
        if value is _EMPTY:
            self.has_value = False
            name = cast(str, name)
            self.value = name.encode("utf-8")
            return

        with _PICKLE_RLOCK:
            self.has_value = True
            name_owner: FunctionType | type
            if getattr(value, "__qualname__", None) not in (None, ""):
                qname = value.__qualname__
                name_owner = value
                real_mod = getattr(value, "__module__", None)
                if real_mod in (None, ""):
                    raise ValueError(f"对象 {value} 所属模块无法找到，因此不能被 pickle")
            else:
                cls = getattr(value, "__class__", _EMPTY)
                if cls is _EMPTY:
                    raise ValueError(f"对象 {value} 的类无法找到，因此不能被 pickle")

                _qname = cast(str | None, getattr(cls, "__qualname__", None))
                if _qname in (None, ""):
                    raise ValueError(f"对象 {value} 的类没有有效的名称，因此不能被 pickle")
                qname = cast(str, _qname)
                name_owner = cast(type, cls)

                real_mod = getattr(cls, "__module__", None)
                if real_mod in (None, ""):
                    raise ValueError(f"对象 {value} 的类所属模块无法找到，因此不能被 pickle")

            real_mod = cast(str | None, real_mod)
            real_name = getattr(name_owner, "__name__", _EMPTY)
            real_qname = getattr(name_owner, "__qualname__", _EMPTY)

            # 获取 pickle 要验证的模块，如果存在则暂时保存
            mod = sys.modules.get(self.module)
            orig_mod = mod if mod is not None else None
            fake_mod = ModuleType(self.module)

            # 逐级生成属性，直到最后一层
            parts = qname.split(".")
            idx = 0
            node = fake_mod
            while idx < len(parts) - 1:
                setattr(node, parts[idx], _DUMMY_CLS())
                node = getattr(node, parts[idx])
                idx += 1
            # 最后一层填充实际值
            setattr(node, parts[-1], name_owner)

            # 对于实际的对象，构建临时环境来欺骗 pickle
            sys_modified = False
            try:
                name_owner.__qualname__ = qname
                name_owner.__name__ = qname.split(".")[-1]
                name_owner.__module__ = self.module
                sys.modules[self.module] = fake_mod
                sys_modified = True
                self.value = pickle.dumps(value)
            finally:
                if sys_modified:
                    if orig_mod is None:
                        del sys.modules[self.module]
                    else:
                        sys.modules[self.module] = orig_mod

                if real_mod is None:
                    if hasattr(name_owner, "__module__"):
                        del name_owner.__module__
                else:
                    name_owner.__module__ = real_mod

                if real_name is _EMPTY:
                    if hasattr(name_owner, "__name__"):
                        del name_owner.__name__
                else:
                    name_owner.__name__ = cast(str, real_name)

                if real_qname is _EMPTY:
                    if hasattr(name_owner, "__qualname__"):
                        del name_owner.__qualname__
                else:
                    name_owner.__qualname__ = cast(str, real_qname)

    def __reduce__(self) -> tuple[Callable, tuple[Any, ...]]:
        try:
            self._serialize(*self._serial_args)
        except Exception as e:
            if self._orig_value is _EMPTY:
                raise pickle.PicklingError(f"Pickle 失败，{e}") from e
            else:
                raise pickle.PicklingError(f"Pickle 对象 {self._orig_value} 失败，{e}") from e
        else:
            return (_deserialize, (self.value, self.has_value, self.module, self.entry))


def _deserialize(value: bytes, has_value: bool, module: str, entry: str | None) -> Any:
    from ._imp import Importer

    with _PICKLE_RLOCK:
        try:
            dir_path = (
                Path(entry).resolve(strict=True).parent.as_posix() if entry is not None else None
            )
            mod = Importer.import_mod(module, dir_path)
            if has_value:
                return pickle.loads(value)
            else:
                return getattr(mod, value.decode("utf-8"))
        except Exception as e:
            raise pickle.UnpicklingError(f"Unpickle 失败，{e}") from e
